# Smoothed indicator function
function Φ(x)
    return 0.5 * x / sqrt(1.0e-7 + x^2) + 0.5
end

# Intermediate computation for choice probabiltiies
function Φ!(Φ_temp, P_temp, U_temp, j, nT)

    for i in 1:nT
        if i == j
            Φ_temp[i] = Φ(U_temp[j] + P_temp[j] - U_temp[nT + 1])
        else
            Φ_temp[i] = Φ(U_temp[j] + P_temp[j] - U_temp[i] - P_temp[i])
        end
    end

end

# Moments to be matched (choice probabilities + location/scale normalizations)
function csw_moments!(K, G, θ, U, obj)

    G .= 0.0
    K .= 0.0

    @unpack C, J, nT, nC, PF, PM, σ2 = obj.γ

    T = zeros(eltype(θ), nT, nT)
    @views T[J:J + 1, :] .= reshape(θ, (2, nT))

    G_skip = 2 * (nT + 1)

    # pre-allocate terms to be looped over
    Φ_temp = zeros(eltype(θ), nT)
    F_temp = zeros(eltype(θ), nT)

    # location and scale normalizations
    @views G[:, 1:G_skip ÷ 2] .= U[:, 1:G_skip ÷ 2]
    @views G[:, G_skip ÷ 2 + 1:G_skip] .= U[:, 1:G_skip ÷ 2] .^ 2 .- σ2

    # individual type
    for (i, ii) in enumerate(vcat(J, J + 1))

        # rows of T and Z correspond to female type, columns correpsond to male type
        @views F_temp .= T[ii, :]

        # spouse type
        for j in 1:nT

            # females
            colf = G_skip + (i - 1) * nT + j
            ccpf = PF[ii, C, j]

            for m in 1:size(U)[1]
                Φ!(Φ_temp, F_temp, @view(U[m, :]), j, nT)
                G[m, colf] = prod(Φ_temp) - ccpf
            end

        end
    end

    # counterfactual: female J to J+1 premium
    ccp = zeros(eltype(θ), nT, 2)
    pp = zeros(eltype(θ), nT + 1, 2)
    TT = zeros(eltype(θ), nT + 1, 2)
    Tu = zeros(eltype(θ), nT + 1, 2)
    transpose!(ccp, @view(PF[J:J + 1, C, :]))
    @views TT[1:nT, 1] .= T[J, :]
    @views TT[1:nT, 2] .= T[J + 1, :]
    col0 = (1:nT) .+ G_skip
    col1 = col0 .+ nT
    for m in 1:size(U)[1]
        @views pp[1:nT, 1] .= G[m, col0]
        @views pp[1:nT, 2] .= G[m, col1]
        @views pp[1:nT, :] .+= ccp
        @views pp[nT + 1, 1] = 1.0 - sum(pp[1:nT, 1])
        @views pp[nT + 1, 2] = 1.0 - sum(pp[1:nT, 2])
        Tu .= TT
        Tu .+= @view(U[m, 1:nT + 1])
        @views K[m] = dot(pp[:, 2], Tu[:, 2]) - dot(pp[:, 1], Tu[:, 1])
    end

end

# As above, for CS using projection
function csw_moments_proj!(K, G, θ, U, obj)

    G .= 0.0
    K .= 0.0

    @unpack C, J, nT, nC, P_CS, σ2 = obj.γ

    T = zeros(eltype(θ), nT, nT)
    @views T[J:J + 1, :] .= reshape(θ, (2, nT))

    G_skip = 2 * (nT + 1)

    # pre-allocate terms to be looped over
    Φ_temp = zeros(eltype(θ), nT)
    F_temp = zeros(eltype(θ), nT)

    # location and scale normalizations
    @views G[:, 1:G_skip ÷ 2] .= U[:, 1:G_skip ÷ 2]
    @views G[:, G_skip ÷ 2 + 1:G_skip] .= U[:, 1:G_skip ÷ 2] .^ 2 .- σ2

    # individual type
    for (i, ii) in enumerate(vcat(J, J + 1))

        # rows of T and Z correspond to female type, columns correpsond to male type
        @views F_temp .= T[ii, :]

        # spouse type
        for j in 1:nT

            # females
            colfl = G_skip + (i - 1) * nT + j
            colfu = colfl + 2 * nT
            ccpfu = P_CS[j, i, 2]
            dccpf = P_CS[j, i, 1] - P_CS[j, i, 2]

            for m in 1:size(U)[1]
                Φ!(Φ_temp, F_temp, @view(U[m, :]), j, nT)
                G[m, colfl] = prod(Φ_temp) - ccpfu
                G[m, colfu] = dccpf - G[m, colfl]
            end

        end
    end

    # counterfactual: female J to J+1 premium
    ccp = zeros(eltype(θ), nT, 2)
    pp = zeros(eltype(θ), nT + 1, 2)
    TT = zeros(eltype(θ), nT + 1, 2)
    Tu = zeros(eltype(θ), nT + 1, 2)
    ccp .= @view(P_CS[:, :, 2])
    @views TT[1:nT, 1] .= T[J, :]
    @views TT[1:nT, 2] .= T[J + 1, :]
    col0 = (1:nT) .+ G_skip
    col1 = col0 .+ nT
    for m in 1:size(U)[1]
        @views pp[1:nT, 1] .= G[m, col0]
        @views pp[1:nT, 2] .= G[m, col1]
        @views pp[1:nT, :] .+= ccp
        @views pp[nT + 1, 1] = 1.0 - sum(pp[1:nT, 1])
        @views pp[nT + 1, 2] = 1.0 - sum(pp[1:nT, 2])
        Tu .= TT
        Tu .+= @view(U[m, 1:nT + 1])
        @views K[m] = dot(pp[:, 2], Tu[:, 2]) - dot(pp[:, 1], Tu[:, 1])
    end

end

# As above, imposing exchangeability
function csw_moments_perm!(K, G, θ, U, obj)

    G .= 0.0
    K .= 0.0

    @unpack C, J, nT, nC, PF, PM, σ2, perms, nperm = obj.γ

    T = zeros(eltype(θ), nT, nT)
    @views T[J:J + 1, :] .= reshape(θ, (2, nT))

    # number of normalizations per cohort
    G_skip = 2 * (nT + 1)

    # pre-allocate terms to be looped over
    Φ_temp = zeros(eltype(θ), nT)
    F_temp = zeros(eltype(θ), nT)
    U_temp = copy(U)
    K_temp = zeros(eltype(θ), nT + 1, 2)

    TT = zeros(eltype(θ), nT + 1, 2)
    Tu = zeros(eltype(θ), nT + 1, 2)
    @views TT[1:nT, 1] .= T[J, :]
    @views TT[1:nT, 2] .= T[J + 1, :]

    for p in 1:nperm

        U_temp .= @view(U[:, perms[p]])

        # location and scale normalizations
        @views G[:, 1:G_skip ÷ 2] .+= U_temp[:, 1:G_skip ÷ 2]
        @views G[:, G_skip ÷ 2 + 1:G_skip] .+= U_temp[:, 1:G_skip ÷ 2] .^ 2 .- σ2

        @inbounds for m in 1:size(U)[1]

            # individual type
            @inbounds for (i, ii) in enumerate(vcat(J, J + 1))

                # rows of T and Z correspond to female type, columns correpsond to male type
                @views F_temp .= T[ii, :]

                # spouse type
                @inbounds for j in 1:nT

                    # females
                    colf = G_skip + (i - 1) * nT + j
                    ccpf = PF[ii, C, j]

                    Φ!(Φ_temp, F_temp, @view(U_temp[m, :]), j, nT)
                    pp = prod(Φ_temp)
                    G[m, colf] += pp - ccpf

                    if ii == J
                        K_temp[j, 1] = pp
                    elseif ii == J + 1
                        K_temp[j, 2] = pp
                    end
                end
            end

            @views K_temp[nT + 1, 1] = 1.0 - sum(K_temp[1:nT, 1])
            @views K_temp[nT + 1, 2] = 1.0 - sum(K_temp[1:nT, 2])
            Tu .= TT
            Tu .+= @view(U_temp[m, :])
            @views K[m] += dot(K_temp[:, 2], Tu[:, 2]) - dot(K_temp[:, 1], Tu[:, 1])

        end

    end

    G ./= nperm
    K ./= nperm

end

function csw_moments_perm_threaded!(K, G, θ, U, obj)

    T = Threads.nthreads()
    Threads.@threads for t = 1:T

        ix0 = round(Int, (t - 1) / T * size(U)[1]) + 1
        ix1 = round(Int, t / T * size(U)[1])

        csw_moments_perm!(@view(obj.H[ix0:ix1, 1]), select_G_from_H(obj, @view(obj.H[ix0:ix1, :])), θ, @view(obj.U[ix0:ix1, :]), obj)

    end

end

# As above, imposing exchangeability for projection inference
function csw_moments_perm_proj!(K, G, θ, U, obj)

    G .= 0.0
    K .= 0.0

    @unpack C, J, nT, nC, P_CS, σ2, perms, nperm = obj.γ

    T = zeros(eltype(θ), nT, nT)
    @views T[J:J + 1, :] .= reshape(θ, (2, nT))

    # number of normalizations per cohort
    G_skip = 2 * (nT + 1)

    # pre-allocate terms to be looped over
    Φ_temp = zeros(eltype(θ), nT)
    F_temp = zeros(eltype(θ), nT)
    U_temp = copy(U)
    K_temp = zeros(eltype(θ), nT + 1, 2)

    TT = zeros(eltype(θ), nT + 1, 2)
    Tu = zeros(eltype(θ), nT + 1, 2)
    @views TT[1:nT, 1] .= T[J, :]
    @views TT[1:nT, 2] .= T[J + 1, :]

    for p in 1:nperm

        U_temp .= @view(U[:, perms[p]])

        # location and scale normalizations
        @views G[:, 1:G_skip ÷ 2] .+= U_temp[:, 1:G_skip ÷ 2]
        @views G[:, G_skip ÷ 2 + 1:G_skip] .+= U_temp[:, 1:G_skip ÷ 2] .^ 2 .- σ2

        @inbounds for m in 1:size(U)[1]

            # individual type
            @inbounds for (i, ii) in enumerate(vcat(J, J + 1))

                # rows of T and Z correspond to female type, columns correpsond to male type
                @views F_temp .= T[ii, :]

                # spouse type
                @inbounds for j in 1:nT

                    # females
                    colfl = G_skip + (i - 1) * nT + j
                    colfu = colfl + 2 * nT
                    ccpfu = P_CS[j, i, 2]
                    ccpfl = P_CS[j, i, 1]

                    Φ!(Φ_temp, F_temp, @view(U_temp[m, :]), j, nT)
                    pp = prod(Φ_temp)
                    G[m, colfl] += pp - ccpfu
                    G[m, colfu] += ccpfl - pp

                    if ii == J
                        K_temp[j, 1] = pp
                    elseif ii == J + 1
                        K_temp[j, 2] = pp
                    end
                end
            end

            @views K_temp[nT + 1, 1] = 1.0 - sum(K_temp[1:nT, 1])
            @views K_temp[nT + 1, 2] = 1.0 - sum(K_temp[1:nT, 2])
            Tu .= TT
            Tu .+= @view(U_temp[m, :])
            @views K[m] += dot(K_temp[:, 2], Tu[:, 2]) - dot(K_temp[:, 1], Tu[:, 1])

        end

    end

    G ./= nperm
    K ./= nperm

end

function csw_moments_perm_proj_threaded!(K, G, θ, U, obj)

    T = Threads.nthreads()
    Threads.@threads for t = 1:T

        ix0 = round(Int, (t - 1) / T * size(U)[1]) + 1
        ix1 = round(Int, t / T * size(U)[1])

        csw_moments_perm_proj!(@view(obj.H[ix0:ix1, 1]), select_G_from_H(obj, @view(obj.H[ix0:ix1, :])), θ, @view(obj.U[ix0:ix1, :]), obj)

    end

end

# Jacobian threaded
function csw_moments_jacobian_threaded!(jac_K, jac_G, θ, U, obj)

    T = Threads.nthreads()
    Threads.@threads for t = 1:T

        ix0 = round(Int, (t - 1) / T * size(U)[1]) + 1
        ix1 = round(Int, t / T * size(U)[1])

        f = (H, ϑ) -> begin
            csw_moments!(@view(H[:, 1]), select_G_from_H(obj, H), ϑ, @view(obj.U[ix0:ix1, :]), obj)
        end

        @views ForwardDiff.jacobian!(
            reshape(obj.jac_h[ix0:ix1, :, :], (ix1-ix0+1) * size(obj.H, 2), obj.l),
            f,
            obj.H[ix0:ix1, :],
            θ,
            ForwardDiff.JacobianConfig(f, obj.H[ix0:ix1, :], θ), Val{true}())

    end

    jac_K .= @view(obj.jac_h[1:size(U)[1], 1, :])

end

function csw_moments_proj_jacobian_threaded!(jac_K, jac_G, θ, U, obj)

    T = Threads.nthreads()
    Threads.@threads for t = 1:T

        ix0 = round(Int, (t - 1) / T * size(U)[1]) + 1
        ix1 = round(Int, t / T * size(U)[1])

        f = (H, ϑ) -> begin
            csw_moments_proj!(@view(H[:, 1]), select_G_from_H(obj, H), ϑ, @view(obj.U[ix0:ix1, :]), obj)
        end

        @views ForwardDiff.jacobian!(
            reshape(obj.jac_h[ix0:ix1, :, :], (ix1-ix0+1) * size(obj.H, 2), obj.l),
            f,
            obj.H[ix0:ix1, :],
            θ,
            ForwardDiff.JacobianConfig(f, obj.H[ix0:ix1, :], θ), Val{true}())

    end

    jac_K .= @view(obj.jac_h[1:size(U)[1], 1, :])

end

function csw_moments_perm_jacobian_threaded!(jac_K, jac_G, θ, U, obj)

    T = Threads.nthreads()
    Threads.@threads for t = 1:T

        ix0 = round(Int, (t - 1) / T * size(U)[1]) + 1
        ix1 = round(Int, t / T * size(U)[1])

        f = (H, ϑ) -> begin
            csw_moments_perm!(@view(H[:, 1]), select_G_from_H(obj, H), ϑ, @view(obj.U[ix0:ix1, :]), obj)
        end

        @views ForwardDiff.jacobian!(
            reshape(obj.jac_h[ix0:ix1, :, :], (ix1-ix0+1) * size(obj.H, 2), obj.l),
            f,
            obj.H[ix0:ix1, :],
            θ,
            ForwardDiff.JacobianConfig(f, obj.H[ix0:ix1, :], θ), Val{true}())

    end

    jac_K .= @view(obj.jac_h[1:size(U)[1], 1, :])

end

function csw_moments_perm_proj_jacobian_threaded!(jac_K, jac_G, θ, U, obj)

    T = Threads.nthreads()
    Threads.@threads for t = 1:T

        ix0 = round(Int, (t - 1) / T * size(U)[1]) + 1
        ix1 = round(Int, t / T * size(U)[1])

        f = (H, ϑ) -> begin
            csw_moments_perm_proj!(@view(H[:, 1]), select_G_from_H(obj, H), ϑ, @view(obj.U[ix0:ix1, :]), obj)
        end

        @views ForwardDiff.jacobian!(
            reshape(obj.jac_h[ix0:ix1, :, :], (ix1-ix0+1) * size(obj.H, 2), obj.l),
            f,
            obj.H[ix0:ix1, :],
            θ,
            ForwardDiff.JacobianConfig(f, obj.H[ix0:ix1, :], θ), Val{true}())

    end

    jac_K .= @view(obj.jac_h[1:size(U)[1], 1, :])

end
