# Codes to replicate bounds on surplus where F can vary but θ is held fixed at CSW's estimates

# Load packages, etc
import Pkg; Pkg.activate("../..")
using KNITRO, ForwardDiff, Calculus, LinearAlgebra, Distributions, Parameters, Random, JLD, RData
include("../../src/boot.jl")        # Load main programs
include("./aux/csw_boot.jl")        # Load model-specific files and programs

# Load CSW data
data = load("./data/probs_white.RData")
PM = data["pmen"]
PF = data["pwomen"]

# Number of types and cohorts
nT, nC = size(PM)[1:2]

# Education level for counterfactual (an integer between 1 and nT - 1)
J = 3

# Cohort (an integer between 1 and nC)
array_value = Base.parse(Int, ENV["SLURM_ARRAY_TASK_ID"])
C = array_value <= nC ? array_value : array_value - nC

println([Threads.nthreads(), array_value, J, C]); flush(stdout)

# Initialize parameters at CSW's estimates
# Rows correspond to female type, columns correspond to male type
T = zeros(nT, nT)
for i in 1:nT
    T[i, :] .= log.(PF[i, C, :] ./ (1.0 - sum(PF[i, C, :])))
end
θ_initial = T[J:J + 1, :][:]

# Pre-allocate to store estimates for outer optimization across different δ
δ_grid = vcat(10.0.^vcat(-2.0:1.0:2.0))
κ = zeros(length(δ_grid), 2, 2)
flag = zeros(length(δ_grid), 2, 2)

# Make ObjectiveBundle
function make_obj(iter)

    if iter == 1

    # Implementation 1: No exchangeability
    # Set up random draws
    M = 50000
    U = quantile.(GeneralizedExtremeValue(-Base.MathConstants.eulergamma, 1, 0), rhalton(M, nT + 1, singleseed = 1234567))

    obj = PsiObjectiveBundleExplicit(
        δ = 0.01,
        find_smallest = true,
        γ = (C = C, J = J, nT = nT, nC = nC, PF = PF, PM = PM, σ2 = Base.MathConstants.pi^2 / 6.0),
        moments! = csw_moments!,
        moments_jacobian! = csw_moments_jacobian_threaded!,
        d = 2 * (nT + 1) + 2 * nT,
        inequality_index = Int64[],
        l = 2 * nT,
        U = U,
        outer_loop_opt = "./opt/csw_outer_loop_settings_cluster.opt",
        inner_loop_opt = "./opt/csw_inner_loop_settings_cluster.opt",
        lower_limit = -1e1)

    elseif iter == 2

    # Implementation 2: Dihedral exchangeability
    # Permutations
    perms = [[1,2,3,4,5,6],[2,3,4,5,6,1],[3,4,5,6,1,2],[4,5,6,1,2,3],[5,6,1,2,3,4],[6,1,2,3,4,5],
    [6,5,4,3,2,1],[5,4,3,2,1,6],[4,3,2,1,6,5],[3,2,1,6,5,4],[2,1,6,5,4,3],[1,6,5,4,3,2]]
    nperm = length(perms)

    # Set up random draws
    M = 10000
    U0 = quantile.(GeneralizedExtremeValue(-Base.MathConstants.eulergamma, 1, 0), rhalton(M, nT + 1, singleseed = 1234567))
    Uperm = copy(U0)
    for p in 2:nperm
        Utemp = copy(U0[:, perms[p]])
        Uperm = vcat(Uperm, Utemp)
    end

    obj = PsiObjectiveBundleExplicit(
        δ = 0.01,
        find_smallest = true,
        γ = (C = C, J = J, nT = nT, nC = nC, PF = PF, PM = PM, σ2 = Base.MathConstants.pi^2 / 6.0, perms, nperm),
        moments! = csw_moments_perm_threaded!,
        moments_jacobian! = csw_moments_perm_jacobian_threaded!,
        d = 2 * (nT + 1) + 2 * nT,
        inequality_index = Int64[],
        l = 2 * nT,
        U = Uperm,
        outer_loop_opt = "./opt/csw_outer_loop_settings_cluster.opt",
        inner_loop_opt = "./opt/csw_inner_loop_settings_cluster.opt",
        lower_limit = -1e1)

    end

    return obj

end

for iter = 1:2

    obj = make_obj(iter)

    # Estimate
    for (i, δ) in enumerate(δ_grid)
        println(δ); flush(stdout)
        obj.δ = δ
        obj.find_smallest = true
        global (κ[i, 1, iter], x, flag[i, 1, iter]) = inner_loop(obj, θ_initial)
        if flag[i, 1, iter] == -102.0
            obj.x = x
            global (flag[i, 1, iter], κ[i, 1, iter]) = inner_loop_KNITRO(obj)
            global κ[i, 1, iter] *= -1.0
        end
        obj.find_smallest = false
        global (κ[i, 2, iter], x, flag[i, 2, iter]) = inner_loop(obj, θ_initial)
        if flag[i, 2, iter] == -102.0
            obj.x = x
            global (flag[i, 2, iter], κ[i, 2, iter]) = inner_loop_KNITRO(obj)
        end
    end

    save(string("./results/csw_inner_",J,"_",array_value,".jld"),"δ_grid", δ_grid, "κ", κ, "flag", flag, "C", C, "J", J)

end
