# Codes to compute estimates and bootstrap CIs for second implementation (with exchangeability)

# Load packages, etc
import Pkg; Pkg.activate("../..")
using KNITRO, ForwardDiff, Calculus, LinearAlgebra, Distributions, Parameters, Random, JLD, RData
include("../../src/boot.jl")        # Load main programs
include("./aux/csw_boot.jl")        # Load model-specific files and programs

# Load CSW data
data = load("./data/probs_white.RData")
PM = data["pmen"]
PF = data["pwomen"]

# Number of types and cohorts
nT, nC = size(PM)[1:2]

# Education level for counterfactual (an integer between 1 and nT - 1)
J = 3

# Cohort (an integer between 1 and nC)
array_value = Base.parse(Int, ENV["SLURM_ARRAY_TASK_ID"])
C = array_value <= nC ? array_value : array_value - nC

# Finding smallest value?
finding_smallest = array_value <= nC

println([Threads.nthreads(), array_value, J, C, finding_smallest]); flush(stdout)

# Initialize parameters at CSW's estimates
# Rows correspond to female type, columns correspond to male type
T = zeros(nT, nT)
for i in 1:nT
    T[i, :] .= log.(PF[i, C, :] ./ (1.0 - sum(PF[i, C, :])))
end
θ_initial = T[J:J + 1, :][:]

# Step 1: Estimate bounds
# Permutations
perms = [[1,2,3,4,5,6],[2,3,4,5,6,1],[3,4,5,6,1,2],[4,5,6,1,2,3],[5,6,1,2,3,4],[6,1,2,3,4,5],
[6,5,4,3,2,1],[5,4,3,2,1,6],[4,3,2,1,6,5],[3,2,1,6,5,4],[2,1,6,5,4,3],[1,6,5,4,3,2]]
nperm = length(perms)

# Set up random draws
M = 10000
U0 = quantile.(GeneralizedExtremeValue(-Base.MathConstants.eulergamma, 1, 0), rhalton(M, nT + 1, singleseed = 1234567))
Uperm = copy(U0)
for p in 2:nperm
    Utemp = copy(U0[:, perms[p]])
    global Uperm = vcat(Uperm, Utemp)
end

# Assemble the ObjectiveBundle
obj = PsiObjectiveBundleExplicit(
    δ = 0.01,
    find_smallest = finding_smallest,
    γ = (C = C, J = J, nT = nT, nC = nC, PF = PF, PM = PM, σ2 = Base.MathConstants.pi^2 / 6.0, perms, nperm),
    moments! = csw_moments_perm_threaded!,
    moments_jacobian! = csw_moments_perm_jacobian_threaded!,
    d = 2 * (nT + 1) + 2 * nT,
    inequality_index = Int64[],
    l = 2 * nT,
    U = Uperm,
    outer_loop_opt = "./opt/csw_outer_loop_settings_cluster.opt",
    inner_loop_opt = "./opt/csw_inner_loop_settings_cluster.opt",
    lower_limit = -1e1,
    use_cached_x = false)

# Lower and upper bounds on parameter space for θ
θ_lower = -12.0 * ones(obj.l)
θ_upper = +12.0 * ones(obj.l)

# Pre-allocate to store estimates for outer optimization across different δ
δ_grid = 10.0.^vcat(-2.0:0.5:2.0)
κ = zeros(length(δ_grid))
flag = zeros(length(δ_grid))
Θ = zeros(obj.l, length(δ_grid))

# Set up for bootstrap computation
B = 1000                    # Number of draws
ix =  [1, 3, 5, 7, 9]       # δ indices to be bootstrapped
c = zeros(length(ix), B)    # Store values across bootstrap draws
v = zeros(length(ix))       # Store variance estimate using Remark 6.3
Θ_boot = zeros(obj.l, B, length(ix))
flag_boot = zeros(length(ix), B)

# # Uncomment this block and comment out the #Estimate block below if additional time is needed for the bootstrap
# x = load(string("./results/csw_main_2_",J,"_",array_value,".jld"))
# c = x["c"]
# flag_boot = x["flag_boot"]
# κ = x["κ"]
# Θ = x["Θ"]
# flag = x["flag"]

# Estimate
θ1 = copy(θ_initial)
for (i, δ) in enumerate(δ_grid)
    println(δ); flush(stdout)
    obj.δ = δ
    (κ_temp, θ1_temp, flag_temp) = outer_loop_multi(obj, θ_lower, θ_upper, θ1, 10, 0.01)
    if i == 1 || (obj.find_smallest && κ_temp < κ[i-1]) || (!obj.find_smallest && κ_temp > κ[i-1])
        global (κ[i], θ1, flag[i]) = (κ_temp, θ1_temp, flag_temp)
        global Θ[:, i] .= θ1
    else
        global (κ[i], θ1, flag[i]) = (κ[i-1], Θ[:, i-1], flag[i-1])
        global Θ[:, i] .= θ1
    end
    save(string("./results/csw_main_2_",J,"_",array_value,".jld"),"δ_grid", δ_grid, "κ", κ, "c", c, "Θ", Θ, "flag", flag, "flag_boot", flag_boot, "C", C, "J", J, "finding_smallest", finding_smallest, "v", v)
    iter = 1
    while iter <= 5
        println([δ, iter]); flush(stdout)
        (κ_temp, θ1_temp, flag_temp) = outer_loop_multi(obj, θ_lower, θ_upper, θ1, 10, 0.01)
        if (obj.find_smallest && κ_temp < κ[i]) || (!obj.find_smallest && κ_temp > κ[i])
            global (κ[i], θ1, flag[i]) = (κ_temp, θ1_temp, flag_temp)
            global Θ[:, i] .= θ1
            iter += 1
        else
            iter += 5
        end
        save(string("./results/csw_main_2_",J,"_",array_value,".jld"),"δ_grid", δ_grid, "κ", κ, "c", c, "Θ", Θ, "flag", flag, "flag_boot", flag_boot, "C", C, "J", J, "finding_smallest", finding_smallest, "v", v)
    end
end

# Step 2: Bootstrap
# Update options
obj.outer_loop_opt = "./opt/csw_outer_loop_settings_cluster_boot.opt"
obj.inner_loop_opt = "./opt/csw_inner_loop_settings_cluster_boot.opt"
obj.use_cached_x = false

# Cutoff parameter
n = sum(data["nwomen"][J: J + 1, C]) / 2
obj.moments!(@view(obj.H[:, 1]), select_G_from_H(obj, obj.H), θ_initial, obj.U, obj)
ε = minimum(diag(cov(obj.H[:, 3:end]))) * sqrt(log(n) / n)

# Generate bootstrap draws by resampling match probability estimates from normal approximation
z, V = csw_generate_bootstrap_draws(obj, data, nT, J, C, B)

# Bootstrap
for i in 1:length(ix)
    δ = δ_grid[ix[i]]
    obj.δ = δ
    println(δ); flush(stdout)
    global θ1 = Θ[:, ix[i]]
    for b in 1:B
        if c[i, b] == 0.0
        println([δ, b]); flush(stdout)
        global (c[i, b], θ2, flag_boot[i, b]) = outer_loop_bootstrap(obj, θ_lower, θ_upper, θ1, z[:, b], ε)
        global Θ_boot[:, b, i] .= θ2
        save(string("./results/csw_main_2_",J,"_",array_value,".jld"),"δ_grid", δ_grid, "κ", κ, "c", c, "Θ", Θ, "flag", flag, "flag_boot", flag_boot, "C", C, "J", J, "finding_smallest", finding_smallest, "v", v)
        end
    end
end

# Variance estimate using Remark 6.3
for i in 1:length(ix)
    δ = δ_grid[ix[i]]
    obj.δ = δ
    println(δ); flush(stdout)
    global θ1 = Θ[:, ix[i]]
    global v[i] = variance_singleton(obj, θ1, V)
end
save(string("./results/csw_main_2_",J,"_",array_value,".jld"),"δ_grid", δ_grid, "κ", κ, "c", c, "Θ", Θ, "flag", flag, "flag_boot", flag_boot, "C", C, "J", J, "finding_smallest", finding_smallest, "v", v)
