# Codes to replicate bounds for first implementation with different ϕ-divergences

# Load packages, etc
import Pkg; Pkg.activate("../..")
using KNITRO, ForwardDiff, Calculus, LinearAlgebra, Distributions, Parameters, Random, JLD, RData
include("../../src/boot.jl")        # Load main programs
include("./aux/csw_boot.jl")        # Load model-specific files and programs

# Load CSW data
data = load("./data/probs_white.RData")
PM = data["pmen"]
PF = data["pwomen"]

# Number of types and cohorts
nT, nC = size(PM)[1:2]

# Education level for counterfactual (an integer between 1 and nT - 1)
J = 3

# Cohort (an integer between 1 and nC)
array_value = Base.parse(Int, ENV["SLURM_ARRAY_TASK_ID"])
C = array_value <= nC ? array_value : array_value - nC

# Finding smallest value?
finding_smallest = array_value <= nC

println([Threads.nthreads(), array_value, J, C, finding_smallest]); flush(stdout)

# Initialize parameters at CSW's estimates
# Rows correspond to female type, columns correspond to male type
T = zeros(nT, nT)
for i in 1:nT
    T[i, :] .= log.(PF[i, C, :] ./ (1.0 - sum(PF[i, C, :])))
end
θ_initial = T[J:J + 1, :][:]

# Set up random draws
M = 50000
U = quantile.(GeneralizedExtremeValue(-Base.MathConstants.eulergamma, 1, 0), rhalton(M, nT + 1, singleseed = 1234567))

# Assemble the ObjectiveBundle
obj = PsiObjectiveBundleExplicit(
    δ = 0.01,
    find_smallest = finding_smallest,
    γ = (C = C, J = J, nT = nT, nC = nC, PF = PF, PM = PM, σ2 = Base.MathConstants.pi^2 / 6.0),
    moments! = csw_moments!,
    moments_jacobian! = csw_moments_jacobian_threaded!,
    d = 2 * (nT + 1) + 2 * nT,
    inequality_index = Int64[],
    l = 2 * nT,
    U = U,
    outer_loop_opt = "./opt/csw_outer_loop_settings_cluster.opt",
    inner_loop_opt = "./opt/csw_inner_loop_settings_cluster.opt",
    lower_limit = -1e1)

# Lower and upper bounds on parameter space for θ
θ_lower = -12.0 * ones(obj.l)
θ_upper = +12.0 * ones(obj.l)

# Pre-allocate to store estimates for outer optimization across different δ
δ_grid = 10.0.^vcat(-2.0:0.5:2.0)
κ = zeros(length(δ_grid), 3)
flag = zeros(length(δ_grid), 3)
Θ = zeros(obj.l, length(δ_grid), 3)

# Estimate
for j in 1:3

    if j == 2
        obj.Psi! = Psi_χ2!
        obj.dPsi! = dPsi_χ2!
        obj.ddPsi! = ddPsi_χ2!
    elseif j == 3
        obj.Psi! = Psi_l4!
        obj.dPsi! = dPsi_l4!
        obj.ddPsi! = ddPsi_l4!
    end

    global θ1 = copy(θ_initial)
    for (i, δ) in enumerate(δ_grid)

        println(j, " ", i, " ", δ); flush(stdout)
        obj.δ = δ
        (κ_temp, θ1_temp, flag_temp) = outer_loop_multi(obj, θ_lower, θ_upper, θ1, 10, 0.01)
        if i == 1 || (obj.find_smallest && κ_temp < κ[i-1, j]) || (!obj.find_smallest && κ_temp > κ[i-1, j])
            global (κ[i, j], θ1, flag[i, j]) = (κ_temp, θ1_temp, flag_temp)
            global Θ[:, i, j] .= θ1
        else
            global (κ[i, j], θ1, flag[i, j]) = (κ[i-1, j], Θ[:, i-1, j], flag[i-1, j])
            global Θ[:, i, j] .= θ1
        end
        save(string("./results/csw_phi_",J,"_",array_value,".jld"),"δ_grid", δ_grid, "κ", κ, "Θ", Θ, "flag", flag, "C", C, "J", J, "finding_smallest", finding_smallest)
        iter = 1
        while iter <= 5
            println(j, " ", i, " ", δ, " ", iter); flush(stdout)
            (κ_temp, θ1_temp, flag_temp) = outer_loop_multi(obj, θ_lower, θ_upper, θ1, 10, 0.01)
            if (obj.find_smallest && κ_temp < κ[i, j]) || (!obj.find_smallest && κ_temp > κ[i, j])
                global (κ[i, j], θ1, flag[i, j]) = (κ_temp, θ1_temp, flag_temp)
                global Θ[:, i, j] .= θ1
                iter += 1
            else
                iter += 5
            end
            save(string("./results/csw_phi_",J,"_",array_value,".jld"),"δ_grid", δ_grid, "κ", κ, "Θ", Θ, "flag", flag, "C", C, "J", J, "finding_smallest", finding_smallest)
        end
    end
end
