# STEP 1: Load and pre-process the data using a subset of the original code

rm(list=ls())


################ AUXILIARY FUNCTIONS ################################################

## logarithm extended by a quadratic below epsl, 
##   with first derivative
##  x should be a number or a vector of numbers

bslog <- function(x, epsl=0.00000001) {
  onex <- 0*x+1
  z <- pmin(x/epsl,onex)
  epslx <- 0*x+epsl
  xcorr <- pmax(epslx,x)
  l <- log(xcorr)-1.5+2.0*z-z*z*0.5
  dl <- 0*l
  beleps <- sum(x < epsl)
  if (beleps > 0) {
    dl[x < epsl] <- 2.0/epsl-z[x < epsl]/epsl
  }
  if (beleps < NROW(x)) { 
    dl[x >= epsl] <- 1.0/x[x >= epsl]
  }
  ## we return
  list(l,dl,beleps)
}


## limits for plots (argument y can be a vector or an array); from below min to above max
myplim <- function(y,eps=0.1) {
  a <- min(y)
  b <- max(y)
  c(a-eps*(b-a),b+eps*(b-a))
}


################################################################################################

require(MASS)
require(lattice)
require(xtable)
require(ggplot2)
require(numDeriv)

##  we take man older by 
manolderby <- 1
yearadj <- "years"
if (manolderby==1) {
  yearadj <- "year"
}
## later we may take expectation over age difference, can we do it consistently on both sides?

## select the race
selectRace <- 1

## level of aggregation
FiveLevels <- T        ## we use the five levels of education 
ThreeLevels <- F
SCwithCG <- F
CGwithCGplus <- F

racename <- "Whites"
raceadjs <- "whites"
raceadj <- "white"


## we only keep people observed at age >= younglim
younglim <-  40
if (selectRace==3) {
  younglim <- 45
}

## we normalize on sigma(normg)=1
if (FiveLevels | CGwithCGplus) {
  normg <- 3
}
if (ThreeLevels) {
  normg <- 2
}

cat("We work with",raceadjs,"\n")
cat(" only those at least",younglim,"when observed\n")

cat("  late marriages reclassified\n")

cat("  woman ", manolderby, yearadj, "younger than husband\n")
if (FiveLevels) {
  cat(" Five degrees of education.\n")
}
if (CGwithCGplus) {
  cat(" Four degrees of education.\n")
}
if (ThreeLevels) {
  cat(" Three degrees of education.\n")
}
cat(" Normalization sigma(",normg,")=1\n")


load(file=paste0(racename,"/probsnonpar",
                selectRace,younglim,".RData"))

if (ThreeLevels) {
  aggregMat <- matrix(0,3,5)
  aggregMat[1:3,1:3] <- diag(3)
  aggregMat[3,4:5] <- 1
  LevelsMessage <- "We aggregate in three levels: \n     HSD, HSG, and SC (with CG and CG+)\n"
  smallnamesgroups <-  c("HSD", "HSG","SC ")
  namesgroups <-  c("High School Dropouts", "High School Graduates","Some College")
  bannergroups <- "       HSD     HSG     SC   \n"
  dirng <- "Three/"
}
if (SCwithCG) {
  aggregMat <- matrix(0,4,5)
  aggregMat[1:3,1:3] <- diag(3)
  aggregMat[3,4] <- 1
  aggregMat[4,5] <- 1
  LevelsMessage <- "We aggregate in four levels:\n    HSD, HSG, SC (with CG) and CG+\n"
  namesgroups <-  c("High School Dropouts", "High School Graduates","Some College","Postgraduate")
  smallnamesgroups <-  c("HSD", "HSG","SC ", "CG+")
  bannergroups <- "       HSD     HSG     SC    CG+ \n"
}
if (CGwithCGplus) {
  aggregMat <- matrix(0,4,5)
  aggregMat[1:4,1:4] <- diag(4)
  aggregMat[4,5] <- 1
  LevelsMessage <- "We aggregate in four levels:\n   HSD, HSG, SC and CG (with CG+)\n"
  namesgroups <-  c("High School Dropouts", "High School Graduates","Some College","College")
  smallnamesgroups <-  c("HSD", "HSG","SC ", "CG")
  bannergroups <- "       HSD     HSG     SC    CG \n"
  dirng <- "Four/"
}
if (FiveLevels) {
  aggregMat <- diag(5)
  LevelsMessage <- "We keep all five levels:\n   HSD, HSG, SC, CG and CG+\n"
  namesgroups <-  c("High School Dropouts", "High School Graduates","Some College",
                      "College Graduate", "Postgraduate")
  smallnamesgroups <-  c("HSD", "HSG","SC ","CG ","CG+")
  bannergroups <- "       HSD     HSG     SC     CG    CG+ \n"
  dirng <- "Five/"
}


## we focus on cohorts of men between bcoh and ecoh
bcoh <- 1940
ecoh <- 2008-younglim-manolderby
cohyears <- (bcoh:ecoh)
ncoh <- ecoh-bcoh+1
cohinds <- seq(ncoh)


## aggregate groups
## Amat(ngnew,ngold)  maps old groups to new groups
aggregateGroups <- function(probs,nb,varp,Amat) {
  ##browser()
  ngnew <- NROW(Amat)
  ngold <- NCOL(Amat)
  ncoh <- NCOL(nb)
  pnew <- array(0,c(ngnew,ncoh,(ngnew+1)))
  vpnew <- array(0,c(ngnew,ncoh,(ngnew+1),(ngnew+1)))
  notsingold <- (2:(ngold+1))
  notsingnew <- (2:(ngnew+1))
  ## numbers of indivs in each new group in each cohort
  newnb <- Amat %*% nb
  for (icoh in 1:ncoh) {
    nbi <- nb[,icoh]
    newnbi <- newnb[,icoh]
    D1new <- diag(1/newnbi)
    ## not single
    p <- probs[,icoh,notsingold]
    Anold <- t(t(Amat)*nbi)
    newp <- (Anold %*% (p %*% t(Amat)))/newnbi
    ## single
    p0 <- probs[,icoh,1]
    newp0 <- (Anold %*% p0)/newnbi
    ## together
    pnew[,icoh,1] <- newp0
    pnew[,icoh,notsingnew] <- newp
    ## now variances
    D2old <- diag(nbi*nbi)
    D2new <- diag(1/(newnbi*newnbi))
    VKnew <- array(0,c(ngold,(ngnew+1),(ngnew+1)))
    Voldi <- varp[,icoh,,]
    for (K in 1:ngold) {
      VKold <- Voldi[K,,]
      VKnew[K,notsingnew,notsingnew] <- 
        (Amat %*% VKold[notsingold,notsingold]) %*% t(Amat)
      VKnew[K,1,notsingnew] <- VKold[1,notsingold] %*% t(Amat)
      VKnew[K,1,1] <- VKold[1,1]
    }
    An2 <- Anold/newnbi
    An2 <- An2*An2
    vpnew[,icoh,1,1] <- An2 %*% VKnew[,1,1]
    vpnew[,icoh,1,notsingnew] <- An2 %*% VKnew[,1,notsingnew]
    vpnew[,icoh,notsingnew,1] <- vpnew[,icoh,1,notsingnew]
    for (i in notsingnew) {
      vpnew[,icoh,notsingnew,i] <- An2 %*% VKnew[,notsingnew,i]
    }
  }
  ## we return
  list(pnew,newnb,vpnew)
}




vcohmen <- cohyears-begcohmen+1
vcohwomen <- cohyears-begcohwomen+1+manolderby

newmen <- aggregateGroups(probsmen,nmen,varpmen,aggregMat)
newwomen <- aggregateGroups(probswomen,nwomen,varpwomen,aggregMat)
probsmen <- newmen[[1]][,vcohmen,]
nmen <- newmen[[2]][,vcohmen]
varpmen <- newmen[[3]][,vcohmen,,]
probswomen <- newwomen[[1]][,vcohwomen,]
nwomen <- newwomen[[2]][,vcohwomen]
varpwomen <- newwomen[[3]][,vcohwomen,,]

## pick proba of single
pmen0 <- probsmen[,,1]
pwomen0 <- probswomen[,,1]
varpmen0 <- varpmen[,,1,1]
varpwomen0 <- varpwomen[,,1,1]

ng <- NROW(pmen0)
ng1 <- ng+1
ng2 <- ng*ng
ngm <- ng-1
ncohm <- ncoh-1

##  now probas of marrying the various types of partners
pmen <- probsmen[,,(2:ng1)]
pwomen <- probswomen[,,(2:ng1)]
## these are 3-dim arrays,
##  first dim=group of person
##  second dim=cohort of person
##  third dim= group of partner

varpmen <- varpmen[,,(2:ng1),(2:ng1)]
varpwomen <- varpwomen[,,(2:ng1),(2:ng1)]

save(pmen, pwomen, nmen, nwomen, varpmen, varpwomen, pmen0, pwomen0, varpmen0, varpwomen0, file = "probs_white.RData")
