# Codes to plot all figures for the CSW example

# Assumes results have been generated by running
# csw_main_1.jl,
# csw_main_2.jl,
# csw_proj_1.jl,
# csw_proj_2.jl,
# csw_inner.jl,
# csw_phi.jl

# Load packages, etc
import Pkg; Pkg.activate("../..")
using KNITRO, ForwardDiff, Calculus, LinearAlgebra, Distributions, Parameters, Random, JLD, RData, Plots, LaTeXStrings, BenchmarkTools, ColorSchemes, StatsBase
include("../../src/boot.jl")        # Load main programs
include("./aux/csw_boot.jl")        # Load model-specific files and programs

# Load CSW data
data = load("./data/probs_white.RData")
PM = data["pmen"]
PF = data["pwomen"]

# Number of types and cohorts
nT, nC = size(PM)[1:2]

# Education level for counterfactual (an integer between 1 and nT - 1)
J = 3

# Level of significance
α = 0.05

# Save figures?
saving = true

# Formatting
labels = ("δ = 0.01", "δ = 0.1", "δ = 1", "δ = 10", "δ = 100")
colors = [palette(:cubehelix, 15)[i] for i in (1, 3, 5, 8, 9, 11)]

# Part 0: compute estimates and CIs based on CSW's replication files
pwomen0 = data["pwomen0"]
varpwomen0 = data["varpwomen0"]
vexpCS = -log.(pwomen0)
sdvCS = sqrt.(varpwomen0) ./ pwomen0
premwomenCS = vexpCS[2:end, :] .- vexpCS[1:end - 1, :]
sdpremwomenCS = sqrt.(sdvCS[2:end, :] .* sdvCS[2:end, :] .+ sdvCS[1:end - 1, :] .* sdvCS[1:end - 1, :])

# Part 1: plot main estimates and confidence intervals
κ_l = zeros(5, nC, 3, 2)
κ_u = zeros(5, nC, 3, 2)

for iter = 1:2

    # Extract estimates and construct CIs
    for array_value in 1:56
        x = load(string("./results/csw_main_",iter,"_",J,"_",array_value,".jld"))
        for i in 1:5
            if array_value <= nC
                κ_l[i, array_value, 1, iter] = x["κ"][[1, 3, 5, 7, 9][i]]
                c = quantile(x["c"][i, (x["flag_boot"][i, :] .∈ [[0.0, -100.0, -101.0, -103.0]]) .* (abs.(x["c"][i, :]) .< 1e10)], 1 - α)
                κ_l[i, array_value, 2, iter] = κ_l[i, array_value, 1, iter] - c
                κ_l[i, array_value, 3, iter] = κ_l[i, array_value, 1, iter] - quantile(Normal(0, 1), 1 - α) * sqrt(x["v"][i])
            else
                κ_u[i, array_value - nC, 1, iter] = x["κ"][[1, 3, 5, 7, 9][i]]
                c = quantile(x["c"][i, (x["flag_boot"][i, :] .∈ [[0.0, -100.0, -101.0, -103.0]]) .* (abs.(x["c"][i, :]) .< 1e10)], α)
                κ_u[i, array_value - nC, 2, iter] = κ_u[i, array_value - nC, 1, iter] - c
                κ_u[i, array_value - nC, 3, iter] = κ_u[i, array_value - nC, 1, iter] + quantile(Normal(0, 1), 1 - α) * sqrt(x["v"][i])
            end
        end
    end

    # Plot 1: smaller δ
    ix = 1:nC
    p1 = plot(ix, premwomenCS[J, :], lw = 4, color = colors[1], label = "δ = 0")
    plot!(p1, ix, premwomenCS[J, :] .+ quantile(Normal(0, 1), 1 - α / 2) .* sdpremwomenCS[J, :], ls = :dot, lw = 3, color = colors[1], label = "")
    plot!(p1, ix, premwomenCS[J, :] .- quantile(Normal(0, 1), 1 - α / 2) .* sdpremwomenCS[J, :], ls = :dot, lw = 3, color = colors[1], label = "")
    for i = 1:2
        plot!(p1, ix, κ_u[i, :, 1, iter], lw = 4, color = colors[i + 1], label = labels[i])
        plot!(p1, ix, κ_l[i, :, 1, iter], lw = 4, color = colors[i + 1], label = "")
        plot!(p1, ix, κ_u[i, :, 2, iter], ls = :dot, lw = 3, color = colors[i + 1], label = "")
        plot!(p1, ix, κ_l[i, :, 2, iter], ls = :dot, lw = 3, color = colors[i + 1], label = "")
    end
    iter == 1 ? ylabel!(p1, "Premium", ylabelfontsize = 14) : nothing
    plot!(p1, ix, 0 .* vcat(ix), lw = 2, linecolor = :black, label = "")
    plot!(p1, xticks = collect(2:2:nC), xtickfontsize = 12, ytickfontsize = 12)
    plot!(p1, ylims = (-1.6, 1.0))
    iter == 1 ? plot!(p1, legend = false) : plot!(p1, legend = :bottomright, legendfontsize = 14)
    saving == true ? savefig(p1, string("csw_",iter,"_",J,"_short.pdf")) : nothing

    # Plot 2: with larger δ
    ix = 1:nC
    p1 = plot(ix, premwomenCS[J, :], lw = 4, color = colors[1], label = "")
    plot!(p1, ix, premwomenCS[J, :] .+ quantile(Normal(0, 1), 1 - α / 2) .* sdpremwomenCS[J, :], ls = :dot, lw = 3, color = colors[1], label = "")
    plot!(p1, ix, premwomenCS[J, :] .- quantile(Normal(0, 1), 1 - α / 2) .* sdpremwomenCS[J, :], ls = :dot, lw = 3, color = colors[1], label = "")
    for i = 1:5
        plot!(p1, ix, κ_u[i, :, 1, iter], lw = 4, color = colors[i + 1], label = i > 2 ? labels[i] : "")
        plot!(p1, ix, κ_l[i, :, 1, iter], lw = 4, color = colors[i + 1], label = "")
        plot!(p1, ix, κ_u[i, :, 2, iter], ls = :dot, lw = 3, color = colors[i + 1], label = "")
        plot!(p1, ix, κ_l[i, :, 2, iter], ls = :dot, lw = 3, color = colors[i + 1], label = "")
    end
    xlabel!(p1, "Cohort", xlabelfontsize = 14)
    iter == 1 ? ylabel!(p1, "Premium", ylabelfontsize = 14) : nothing
    plot!(p1, ix, 0 .* vcat(ix), lw = 2, linecolor = :black, label = "")
    plot!(p1, xticks = collect(2:2:nC), xtickfontsize = 12, ytickfontsize = 12)
    plot!(p1, ylims = (-7.0, 6.0))
    iter == 1 ? plot!(p1, legend = false) : plot!(p1, legend = :bottomright, legendfontsize = 14)
    saving == true ? savefig(p1, string("csw_",iter,"_",J,"_full.pdf")) : nothing

end

# Part 2: plot values of inner loop at CSW's estimates
for iter = 1:2

    # Extract estimates
    κ_l_inner = zeros(5, nC, 2)
    κ_u_inner = zeros(5, nC, 2)
    for array_value in 1:nC
        x = load(string("./results/csw_inner_",J,"_",array_value,".jld"))
        κ_l_inner[:, array_value, iter] = x["κ"][:, 1, iter]
        κ_u_inner[:, array_value, iter] = x["κ"][:, 2, iter]
    end

    ix = 1:nC
    p1 = plot(ix, premwomenCS[J, :], lw = 4, color = colors[1], ls = :dot, label = "δ = 0")
    for i = 1:5
        plot!(p1, ix, κ_l_inner[i, :, iter], lw = 4, color = colors[i + 1], label = labels[i])
        plot!(p1, ix, κ_u_inner[i, :, iter], lw = 4, color = colors[i + 1], label = "")
    end
    plot!(p1, xlabel = "Cohort",xlabelfontsize = 14)
    plot!(p1, xticks = collect(2:2:nC), xtickfontsize = 12, ytickfontsize = 12)
    plot!(p1, ix, 0 .* vcat(ix), lw = 2, linecolor = :black, label = "")
    plot!(p1, ylims = (-0.7, 0.5))
    iter == 1 ? plot!(p1, legend = false, ylabel = "Premium", ylabelfontsize = 14) : plot!(p1, legend = :bottomright, legendfontsize = 12)
    saving == true ? savefig(p1, string("csw_inner_",iter,"_",J,".pdf")) : nothing

end

# Part 3: Plot least favorable distributions
function make_obj(iter, C)

    if iter == 1

    # Implementation 1: No exchangeability
    # Set up random draws
    M = 50000
    U = quantile.(GeneralizedExtremeValue(-Base.MathConstants.eulergamma, 1, 0), rhalton(M, nT + 1, singleseed = 1234567))

    obj = PsiObjectiveBundleExplicit(
        δ = 0.01,
        find_smallest = true,
        γ = (C = C, J = J, nT = nT, nC = nC, PF = PF, PM = PM, σ2 = Base.MathConstants.pi^2 / 6.0),
        moments! = csw_moments!,
        moments_jacobian! = csw_moments_jacobian_threaded!,
        d = 2 * (nT + 1) + 2 * nT,
        inequality_index = Int64[],
        l = 2 * nT,
        U = U,
        outer_loop_opt = "./opt/csw_outer_loop_settings_cluster.opt",
        inner_loop_opt = "./opt/csw_inner_loop_settings_cluster.opt",
        lower_limit = -1e1)

    elseif iter == 2

    # Implementation 2: Dihedral exchangeability
    # Permutations
    perms = [[1,2,3,4,5,6],[2,3,4,5,6,1],[3,4,5,6,1,2],[4,5,6,1,2,3],[5,6,1,2,3,4],[6,1,2,3,4,5],
    [6,5,4,3,2,1],[5,4,3,2,1,6],[4,3,2,1,6,5],[3,2,1,6,5,4],[2,1,6,5,4,3],[1,6,5,4,3,2]]
    nperm = length(perms)

    # Set up random draws
    M = 10000
    U0 = quantile.(GeneralizedExtremeValue(-Base.MathConstants.eulergamma, 1, 0), rhalton(M, nT + 1, singleseed = 1234567))
    Uperm = copy(U0)
    for p in 2:nperm
        Utemp = copy(U0[:, perms[p]])
        Uperm = vcat(Uperm, Utemp)
    end

    obj = PsiObjectiveBundleExplicit(
        δ = 0.01,
        find_smallest = true,
        γ = (C = C, J = J, nT = nT, nC = nC, PF = PF, PM = PM, σ2 = Base.MathConstants.pi^2 / 6.0, perms, nperm),
        moments! = csw_moments_perm_threaded!,
        moments_jacobian! = csw_moments_perm_jacobian_threaded!,
        d = 2 * (nT + 1) + 2 * nT,
        inequality_index = Int64[],
        l = 2 * nT,
        U = Uperm,
        outer_loop_opt = "./opt/csw_outer_loop_settings_cluster.opt",
        inner_loop_opt = "./opt/csw_inner_loop_settings_cluster.opt",
        lower_limit = -1e1)

    end

    return obj

end

# Correlation matrix under least favorable distributions
function Ρ_lfd(obj)
    Ρ = obj.U' * (obj.arg1 .* obj.U) / (obj.M * obj.γ.σ2)
    return Ρ
end

# Maximum upper triangular element
function max_upper(M)
    m = -Inf
    for i in 1:size(M)[1]
        for j in i+1:size(M)[2]
            m = abs(M[i, j]) >= m ? abs(M[i, j]) : m
        end
    end
    mm = M[abs.(M) .== m]
    return mm[1]
end

# Pre-allocate space to store correlations, etc
Ρ_mat = zeros(5, nC, 2, 2)
S_mat = zeros(5, nC, 2, 2)
H_mat = zeros(5, nC, 2, 2)

for iter = 1:2

    for c in 1:nC

        obj = make_obj(iter, c)

        for k = 1:2

            obj.find_smallest = k == 1 ? true : false
            x = k == 1 ? load(string("./results/csw_main_",iter,"_",J,"_",c,".jld")) : load(string("./results/csw_main_",iter,"_",J,"_",nC + c,".jld"))

            # Loop through and compute LFD to find maximal correlation, etc
            for (i,j) in enumerate((1, 3, 5, 7, 9))
                obj.δ = x["δ_grid"][j]
                inner_loop(obj, x["Θ"][:, j])
                Ρ = Ρ_lfd(obj)
                Ρ_mat[i, c, k, iter] = max_upper(Ρ)
                S_mat[i, c, k, iter] = maximum(abs.([mean(obj.H[:, i]) for i in 15:obj.d + 2]))
                H_mat[i, c, k, iter] = obj.x[1]
            end

            # Plot LFDs in a given problem
            if c == 1 && k == 2

                for h = 1:nT+1
                    p1 = plot(x -> exp(-exp(-x-Base.MathConstants.eulergamma)), ls = :dot, lw = 5, color = colors[1], label = "δ = 0")
                    for (i,j) in enumerate((1, 3, 5, 7, 9))
                        obj.δ = x["δ_grid"][j]
                        inner_loop(obj, x["Θ"][:, j])
                        i > 1 ? plot!(p1, quantile(obj.U[:, h], Weights(obj.arg1), 0.0:0.01:1.0), 0.0:0.01:1.0, lw = 4, color = colors[i + 1], label = labels[i]) : nothing
                    end
                    plot!(p1, xlims = (-2.2, 3.2))
                    plot!(p1, xtickfontsize = 16, ytickfontsize = 16)
                    h == 1 ? plot!(p1, legend = :bottomright, legendfontsize = 16) : plot!(p1, legend = false)
                    saving == true ? savefig(p1, string("csw_qq_",iter,"_",k,"_",h-1,".pdf")) : nothing
                end

            end

        end
    end
end

Ρ_bar = round.(mean(Ρ_mat, dims = 2), digits = 3)
S_bar = round.(mean(maximum(S_mat, dims = 3), dims = 2), digits = 3)

for iter = 1:2
    x = load(string("./results/csw_main_1_",J,"_",1,".jld"))
    println()
    if iter == 1
        println("without exchangeability")
    elseif iter == 2
        println("with exchangeability")
    end
    for (i,j) in enumerate((1, 3, 5, 7, 9))
        println(string(x["δ_grid"][j])," & ",Ρ_bar[i, 1, 1, iter]," & ",Ρ_bar[i, 1, 2, iter]," & ",S_bar[i, 1, 1, iter]," \\\\")
    end
    println()
end

# Part 4: Summarize results with different phi functions
κ_l_phi = zeros(5, nC, 3)
κ_u_phi = zeros(5, nC, 3)

for c in 1:2 * nC
    if c <= nC
        x = load(string("./results/csw_phi_",J,"_",c,".jld"))
        κ_l_phi[:, c, :] .= reshape(x["κ"][vcat(1,3,5,7,9), :, :], (5, 3))
    else
        x = load(string("./results/csw_phi_",J,"_",c,".jld"))
        κ_u_phi[:, c - nC, :] .= reshape(x["κ"][vcat(1,3,5,7,9), :, :], (5, 3))
    end
end

println()
println("width of χ² bounds relative to hybrid bounds")
println(round.(mean((κ_u_phi[:, :, 2] - κ_l_phi[:, :, 2]) ./ (κ_u_phi[:, :, 1] - κ_l_phi[:, :, 1]), dims = 2), digits = 2))
println(round.(median((κ_u_phi[:, :, 2] - κ_l_phi[:, :, 2]) ./ (κ_u_phi[:, :, 1] - κ_l_phi[:, :, 1]), dims = 2), digits = 2))
println()
println("width of L⁴ bounds relative to hybrid bounds")
println(round.(mean((κ_u_phi[:, :, 3] - κ_l_phi[:, :, 3]) ./ (κ_u_phi[:, :, 1] - κ_l_phi[:, :, 1]), dims = 2), digits = 2))
println(round.(median((κ_u_phi[:, :, 3] - κ_l_phi[:, :, 3]) ./ (κ_u_phi[:, :, 1] - κ_l_phi[:, :, 1]), dims = 2), digits = 2))
println()

# Part 6: Computation times
C = 1
T = zeros(nT, nT)
for i in 1:nT
    T[i, :] .= log.(PF[i, C, :] ./ (1.0 - sum(PF[i, C, :])))
end
θ_initial = T[J:J + 1, :][:]
println()
println("computation times")
for iter = 1:2
    for (i, δ) in enumerate(vcat(0.01, 0.1, 1.0, 10.0, 100.0))
        global obj = make_obj(iter, C)
        obj.find_smallest = true
        obj.use_cached_x = false
        inner_loop(obj, θ_initial)
        iter == 1 ? println(string("without exchangeability, δ = ", δ)) : println(string("with exchangeability, δ = ", δ))
        obj.δ = δ
        @btime inner_loop_KNITRO(obj);
    end
    println()
end

# Part 7: Projection CIs
κ_l_proj = zeros(5, nC, 2)
κ_u_proj = zeros(5, nC, 2)

for iter = 1:2

    # Extract estimates and construct CIs
    for array_value in 1:56
        x = load(string("./results/csw_proj_",iter,"_",J,"_",array_value,".jld"))
        for i in 1:5
            if array_value <= nC
                κ_l_proj[i, array_value, iter] = x["κ"][[1, 3, 5, 7, 9][i]]
            else
                κ_u_proj[i, array_value - nC, iter] = x["κ"][[1, 3, 5, 7, 9][i]]
            end
        end
    end

    ix = 1:nC
    p1 = plot(ix, premwomenCS[J, :], lw = 4, color = colors[1], label = (iter == 1 ? "δ = 0" : ""), ls = :dot)
    for i = 1:5
        plot!(p1, ix, κ_u_proj[i, :, iter], lw = 4, color = colors[i + 1], label = ((iter == 2 && i > 1) || (iter == 1 && i <= 1) ? labels[i] : ""))
        plot!(p1, ix, κ_l_proj[i, :, iter], lw = 4, color = colors[i + 1], label = "")
    end
    xlabel!(p1, "Cohort", xlabelfontsize = 14)
    iter == 1 ? ylabel!(p1, "Premium", ylabelfontsize = 14) : nothing
    plot!(p1, ix, 0 .* vcat(ix), lw = 2, linecolor = :black, label = "")
    plot!(p1, xticks = collect(2:2:nC), xtickfontsize = 12, ytickfontsize = 12)
    plot!(p1, ylims = (-11, 10))
    plot!(p1, legend = :bottomright, legendfontsize = 12)
    saving == true ? savefig(p1, string("csw_",iter,"_",J,"_proj.pdf")) : nothing

end
