# Ancillary programs for Rust (1987) example

# Construct transition matrices as in Rust (1987)
function make_transition_matrices(nx, mil)

    p = maximum(mil)
    θγ = zeros(p)
    for i in 1:p
        θγ[i] = mean(mil .== i - 1)
    end
    M00 = transition_matrix(nx, θγ)

    M10 = zeros(nx, nx)
    M10[:, 1] .= 1.0

    return M00, M10

end

# Multinomial transition as in Rust (1987)
function transition_matrix(nx, θγ)

    M = zeros(nx, nx)
    p = length(θγ)

    for i in 1:nx - p
        M[i, i + 0: i + p - 1] = θγ
        M[i, i + p] = 1.0 - sum(θγ)
    end

    for i in 1:p - 1
        M[nx - p + i, nx - p + i:nx - 1] = θγ[1:p - i]
        M[nx - p + i, nx] = 1.0 - sum(θγ[1:p - i])
    end

    M[nx, nx] = 1.0

    return M

end

# Linear payoff specification of payoffs as in Rust (1987), Tables IX and X
# NOTE: payoffs are u(x, 0, θ) - u(x, 1, θ) and incurred under decision 0
function payoffs_linear(θ, nx)

    π = zeros(eltype(θ), nx)

    for x in 1:nx
        π[x] += θ[1] + θ[2] * x
    end

    return π

end

function dpayoffs_linear(θ, x)

    return [1.0, 1.0 * x]

end

# Compute fixed point under type-I extreme value specification
# NOTE: there is a location shift in the distribution, so the Euler-Mascheroni constant does not need to be added
function fixed_point(π0, β, M0, M1)

    max_iter = 1000000
    tol = 1e-8

    V0 = π0 ./ (1.0 - β)
    V1 = V0 .+ 2 * tol

    t0 = similar(V0)
    t1 = similar(V1)

    iter = 1
    while iter <= max_iter && maximum(abs.(V0 - V1)) > tol

        V0 .= V1
        mul!(t0, M0, V0, β, 0.0)
        t0 .+= π0
        mul!(t1, M1, V0, β, 0.0)

        for x in eachindex(V1)
            V1[x] = log_sum_exp(t0[x], t1[x])
        end

        iter += 1

    end

    return V1

end

# use numerically sound log-sum-exp
function log_sum_exp(a, b)

    m = max(a, b)
    l = log(exp(a - m) + exp(b - m))
    return l + m

end

# Compute CCPs assuming indepdent type-I extreme value
function ccps_logit(π0, β, M0, M1, v)

    P = similar(π0)

    for x in 1:size(π0)[1]
        @views P[x] = 1.0 / (1.0 + exp(π0[x] + β * dot(M0[x, :] - M1[x, :],v)))
    end

    return P

end

# Compute the MLE
function rust_mle(data, nx, β, M0, M1)

    function f(θ)
        return rust_log_likelihood(θ, data, nx, β, M0, M1)
    end

    results = optimize(f, zeros(2), Newton(), Optim.Options(show_trace = true, g_tol = 1e-7); autodiff = :forward)
    θ_mle = Optim.minimizer(results)
    H = ForwardDiff.hessian(f, θ_mle)

    return θ_mle, H

end

# Compute the negative log likelihood
function rust_log_likelihood(θ, data, nx, β, M0, M1)

    # profits
    π = payoffs_linear(θ, nx)

    # value function
    v = fixed_point(π, β, M0, M1)

    # ccps
    P = ccps_logit(π, β, M0, M1, v)

    l = zero(eltype(θ))
    for i in eachindex(data["dtx"])
        s = data["dtx"][i]
        d = data["dtc"][i]
        if d == 1
            l += log(P[s])
        elseif d == 0
            l += log(1.0 - P[s])
        end
    end

    return -l

end
