# Smoothed indicator function
function Φ(x)
    return 0.5 * x / sqrt(1.0e-7 + x^2) + 0.5
end

# Derivative of smoothed indicator function for Jacobian
function dΦ(x)
    return 5.0e-8 / (sqrt(1.0e-7 + x^2)^3)
end

# Evaluate moments and counterfactual in-place
function rust_moments!(K, G, θ, U, obj)

    G .= 0.0
    K .= 0.0

    @unpack nx, β, M00, M10, M01, M11, ccps, G_skip, θ_skip, state, σ2, trim, τ = obj.γ

    ccps_trimmed = max.(ccps, trim)

    # unpack θ
    v0 = @view θ[θ_skip + 0 * nx + 1:θ_skip + 1 * nx]
    v1 = @view θ[θ_skip + 1 * nx + 1:θ_skip + 2 * nx]

    # payoffs
    π00 = payoffs_linear(@view(θ[1:2]), nx)
    π01 = payoffs_linear(@view(θ[1:2]) .* [1.0, τ], nx)

    # moments: mean-zero normalization (add any other normalizations here and adjust G_skip)
    G[:, 1:2] .= U
    G[:, 3:4] .= U.^2 .- σ2

    # moments: CCPs and value functions
    for x in 1:nx

        @views EV00 = π00[x] + β * dot(M00[x, :], v0)
        @views EV10 = β * dot(M10[x, :], v0)
        @views EV01 = π01[x] + β * dot(M01[x, :], v1)
        @views EV11 = β * dot(M11[x, :], v1)

        @inbounds for m in 1:size(U)[1]

            v00 = EV00 + U[m, 1]
            v10 = EV10 + U[m, 2]
            v01 = EV01 + U[m, 1]
            v11 = EV11 + U[m, 2]

            p0 = Φ(v10 - v00)
            p1 = Φ(v11 - v01)

            # moments for CCPs
            G[m, G_skip + 0 * nx + x] = p0 - ccps_trimmed[x]

            # moments for value functions
            G[m, G_skip + 1 * nx + x] = p0 * v10 + (1.0 - p0) * v00 - v0[x]
            G[m, G_skip + 2 * nx + x] = p1 * v11 + (1.0 - p1) * v01 - v1[x]

            # default counterfactual is CCP in a particular state
            if x == state
                K[m] = p1
            end

        end
    end

end

# Evaluate Jacobian of moments and counterfactual in-place
function rust_moments_jacobian!(jac_K, jac_G, θ, U, obj)

    @unpack nx, β, M00, M10, M01, M11, ccps, G_skip, θ_skip, state, τ = obj.γ

    jac_K .= 0.0

    # unpack θ
    v0 = @view θ[θ_skip + 0 * nx + 1:θ_skip + 1 * nx]
    v1 = @view θ[θ_skip + 1 * nx + 1:θ_skip + 2 * nx]

    # payoffs
    π00 = payoffs_linear(@view(θ[1:2]), nx)
    π01 = payoffs_linear(@view(θ[1:2]) .* [1.0, τ], nx)

    # moments: CCPs and value functions
    for x in 1:nx

        @views EV00 = π00[x] + β * dot(M00[x, :], v0)
        @views EV10 = β * dot(M10[x, :], v0)
        @views EV01 = π01[x] + β * dot(M01[x, :], v1)
        @views EV11 = β * dot(M11[x, :], v1)

        dπ0 = dpayoffs_linear(θ, x)
        dπ1 = dpayoffs_linear(θ, x) .* [1.0, τ]

        @inbounds for m in 1:size(U)[1]

            v00 = EV00 + U[m, 1]
            v10 = EV10 + U[m, 2]
            v01 = EV01 + U[m, 1]
            v11 = EV11 + U[m, 2]

            p0 = Φ(v10 - v00)
            p1 = Φ(v11 - v01)
            dp0 = dΦ(v10 - v00)
            dp1 = dΦ(v11 - v01)

            t0 = dp0 * (v10 - v00) + p0
            t1 = dp1 * (v11 - v01) + p1

            # G[m, G_skip + 0 * nx + x] = p0 - ccps[x]
            jac_G[m, G_skip + 0 * nx + x, 1:θ_skip] .= dp0 .* -dπ0
            for y in 1:nx
                jac_G[m, G_skip + 0 * nx + x, θ_skip + y] = dp0 * β * (M10[x, y] - M00[x, y])
            end

            # G[m, G_skip + 1 * nx + x] = p0 * v10 + (1.0 - p0) * v00 - v0[x]
            jac_G[m, G_skip + 1 * nx + x, 1:θ_skip] .= (1.0 - t0) .* dπ0
            jac_G[m, G_skip + 1 * nx + x, θ_skip + 0 * nx + x] = 0.0
            for y in 1:nx
                if M10[x, y] > 0.0 || M00[x, y] > 0.0
                    jac_G[m, G_skip + 1 * nx + x, θ_skip + 0 * nx + y] = t0 * β * (M10[x, y] - M00[x, y]) + β * M00[x, y]
                end
            end
            jac_G[m, G_skip + 1 * nx + x, θ_skip + 0 * nx + x] -= 1.0

            # G[m, G_skip + 2 * nx + x] = p1 * v11 + (1.0 - p1) * v01) - v1[x]
            jac_G[m, G_skip + 2 * nx + x, 1:θ_skip] .= (1.0 - t1) .* dπ1
            jac_G[m, G_skip + 2 * nx + x, θ_skip + 1 * nx + x] = 0.0
            for y in 1:nx
                if M11[x, y] > 0.0 || M01[x, y] > 0.0
                    jac_G[m, G_skip + 2 * nx + x, θ_skip + 1 * nx + y] = t1 * β * (M11[x, y] - M01[x, y]) + β * M01[x, y]
                end
            end
            jac_G[m, G_skip + 2 * nx + x, θ_skip + 1 * nx + x] -= 1.0

            # default counterfactual is CCP in a particular state
            if x == state
                jac_K[m, 1:θ_skip] .= dp1 .* -dπ1
                for y in 1:nx
                    jac_K[m, θ_skip + nx + y] = dp1 * β * (M11[x, y] - M01[x, y])
                end
            end

        end
    end

end

# Parallelize Jacobian of moments
function rust_moments_jacobian_threaded!(jac_K, jac_G, θ, U, obj)

    T = Threads.nthreads()
    Threads.@threads for t = 1:T
        ix0 = round(Int, (t - 1) / T * size(U)[1]) + 1
        ix1 = round(Int, t / T * size(U)[1])
        rust_moments_jacobian!(@view(jac_K[ix0:ix1, :]), @view(jac_G[ix0:ix1, :, :]), θ, @view(U[ix0:ix1, :]), obj)
    end

end

# Update rust_moments! with new counterfactual
function rust_moments_welfare!(K, G, θ, U, obj)

    # evaluate moments as above
    rust_moments!(K, G, θ, U, obj)

    # Overwrite counterfactual with welfare
    @unpack nx, θ_skip, T0 = obj.γ
    v0 = @view θ[θ_skip + 0 * nx + 1:θ_skip + 1 * nx]
    v1 = @view θ[θ_skip + 1 * nx + 1:θ_skip + 2 * nx]

    @inbounds for m in 1:size(U)[1]
        K[m] = dot(T0, v1) - dot(T0, v0)
    end

end

# Update Jacobian accordingly
function rust_moments_welfare_jacobian!(jac_K, jac_G, θ, U, obj)

    # Evaluate Jacobian as above
    rust_moments_jacobian!(jac_K, jac_G, θ, U, obj)

    # Overwrite counterfactual Jacobian with welfare
    @unpack nx, θ_skip, T0 = obj.γ

    @inbounds for m in 1:size(U)[1]
        jac_K[m, θ_skip + 0 * nx + 1:θ_skip + 1 * nx] .-= T0
        jac_K[m, θ_skip + 1 * nx + 1:θ_skip + 2 * nx] .+= T0
    end

end

# Parallelize Jacobian of moments
function rust_moments_welfare_jacobian_threaded!(jac_K, jac_G, θ, U, obj)

    T = Threads.nthreads()
    Threads.@threads for t = 1:T
        ix0 = round(Int, (t - 1) / T * size(U)[1]) + 1
        ix1 = round(Int, t / T * size(U)[1])
        rust_moments_welfare_jacobian!(@view(jac_K[ix0:ix1, :]), @view(jac_G[ix0:ix1, :, :]), θ, @view(U[ix0:ix1, :]), obj)
    end

end

# Evaluate moments and counterfactual in-place
function rust_moments_trim!(K, G, θ, U, obj)

    G .= 0.0
    K .= 0.0

    @unpack nx, β, M00, M10, M01, M11, ccps, G_skip, θ_skip, state, σ2, nt, τ = obj.γ

    # unpack θ
    v0 = @view θ[θ_skip + 0 * nx + 1:θ_skip + 1 * nx]
    v1 = @view θ[θ_skip + 1 * nx + 1:θ_skip + 2 * nx]

    # payoffs
    π00 = payoffs_linear(@view(θ[1:2]), nx)
    π01 = payoffs_linear(@view(θ[1:2]) .* [1.0, τ], nx)

    # moments: mean-zero normalization (add any other normalizations here and adjust G_skip)
    G[:, 1:2] .= U
    G[:, 3:4] .= U.^2 .- σ2

    # moments: CCPs and value functions
    for x in 1:nx

        @views EV00 = π00[x] + β * dot(M00[x, :], v0)
        @views EV10 = β * dot(M10[x, :], v0)
        @views EV01 = π01[x] + β * dot(M01[x, :], v1)
        @views EV11 = β * dot(M11[x, :], v1)

        @inbounds for m in 1:size(U)[1]

            v00 = EV00 + U[m, 1]
            v10 = EV10 + U[m, 2]
            v01 = EV01 + U[m, 1]
            v11 = EV11 + U[m, 2]

            p0 = Φ(v10 - v00)
            p1 = Φ(v11 - v01)

            # moments for CCPs
            if x > nt
                G[m, G_skip + x - nt] = p0 - ccps[x]
            end

            # moments for value functions
            G[m, G_skip + 1 * nx + x - nt] = p0 * v10 + (1.0 - p0) * v00 - v0[x]
            G[m, G_skip + 2 * nx + x - nt] = p1 * v11 + (1.0 - p1) * v01 - v1[x]

            # default counterfactual is CCP in a particular state
            if x == state
                K[m] = p1
            end

        end
    end

end

# Evaluate Jacobian of moments and counterfactual in-place
function rust_moments_trim_jacobian!(jac_K, jac_G, θ, U, obj)

    @unpack nx, β, M00, M10, M01, M11, ccps, G_skip, θ_skip, state, nt, τ = obj.γ

    jac_K .= 0.0

    # unpack θ
    v0 = @view θ[θ_skip + 0 * nx + 1:θ_skip + 1 * nx]
    v1 = @view θ[θ_skip + 1 * nx + 1:θ_skip + 2 * nx]

    # payoffs
    π00 = payoffs_linear(@view(θ[1:2]), nx)
    π01 = payoffs_linear(@view(θ[1:2]) .* [1.0, τ], nx)

    # moments: CCPs and value functions
    for x in 1:nx

        @views EV00 = π00[x] + β * dot(M00[x, :], v0)
        @views EV10 = β * dot(M10[x, :], v0)
        @views EV01 = π01[x] + β * dot(M01[x, :], v1)
        @views EV11 = β * dot(M11[x, :], v1)

        dπ0 = dpayoffs_linear(θ, x)
        dπ1 = dpayoffs_linear(θ, x) .* [1.0, τ]

        @inbounds for m in 1:size(U)[1]

            v00 = EV00 + U[m, 1]
            v10 = EV10 + U[m, 2]
            v01 = EV01 + U[m, 1]
            v11 = EV11 + U[m, 2]

            p0 = Φ(v10 - v00)
            p1 = Φ(v11 - v01)
            dp0 = dΦ(v10 - v00)
            dp1 = dΦ(v11 - v01)

            t0 = dp0 * (v10 - v00) + p0
            t1 = dp1 * (v11 - v01) + p1

            # G[m, G_skip + x - nt] = p0 - ccps[x]
            if x > nt
                jac_G[m, G_skip + 0 * nx + x - nt, 1:θ_skip] .= dp0 .* -dπ0
                for y in 1:nx
                    jac_G[m, G_skip + 0 * nx + x - nt, θ_skip + y] = dp0 * β * (M10[x, y] - M00[x, y])
                end
            end

            # G[m, G_skip + 1 * nx + x - nt] = p0 * v10 + (1.0 - p0) * v00 - v0[x]
            jac_G[m, G_skip + 1 * nx + x - nt, 1:θ_skip] .= (1.0 - t0) .* dπ0
            jac_G[m, G_skip + 1 * nx + x - nt, θ_skip + 0 * nx + x] = 0.0
            for y in 1:nx
                if M10[x, y] > 0.0 || M00[x, y] > 0.0
                    jac_G[m, G_skip + 1 * nx + x - nt, θ_skip + 0 * nx + y] = t0 * β * (M10[x, y] - M00[x, y]) + β * M00[x, y]
                end
            end
            jac_G[m, G_skip + 1 * nx + x - nt, θ_skip + 0 * nx + x] -= 1.0

            # G[m, G_skip + 2 * nx + x - nt] = p1 * v11 + (1.0 - p1) * v01) - v1[x]
            jac_G[m, G_skip + 2 * nx + x - nt, 1:θ_skip] .= (1.0 - t1) .* dπ1
            jac_G[m, G_skip + 2 * nx + x - nt, θ_skip + 1 * nx + x] = 0.0
            for y in 1:nx
                if M11[x, y] > 0.0 || M01[x, y] > 0.0
                    jac_G[m, G_skip + 2 * nx + x - nt, θ_skip + 1 * nx + y] = t1 * β * (M11[x, y] - M01[x, y]) + β * M01[x, y]
                end
            end
            jac_G[m, G_skip + 2 * nx + x - nt, θ_skip + 1 * nx + x] -= 1.0

            # default counterfactual is CCP in a particular state
            if x == state
                jac_K[m, 1:θ_skip] .= dp1 .* -dπ1
                for y in 1:nx
                    jac_K[m, θ_skip + nx + y] = dp1 * β * (M11[x, y] - M01[x, y])
                end
            end

        end
    end

end

# Update rust_moments_trim! with new counterfactual
function rust_moments_trim_welfare!(K, G, θ, U, obj)

    # evaluate moments as above
    rust_moments_trim!(K, G, θ, U, obj)

    # Overwrite counterfactual with welfare
    @unpack nx, θ_skip, T0 = obj.γ
    v0 = @view θ[θ_skip + 0 * nx + 1:θ_skip + 1 * nx]
    v1 = @view θ[θ_skip + 1 * nx + 1:θ_skip + 2 * nx]

    @inbounds for m in 1:size(U)[1]
        K[m] = dot(T0, v1) - dot(T0, v0)
    end

end

# Update Jacobian accordingly
function rust_moments_trim_welfare_jacobian!(jac_K, jac_G, θ, U, obj)

    # Evaluate Jacobian as above
    rust_moments_trim_jacobian!(jac_K, jac_G, θ, U, obj)

    # Overwrite counterfactual Jacobian with welfare
    @unpack nx, θ_skip, T0 = obj.γ

    @inbounds for m in 1:size(U)[1]
        jac_K[m, θ_skip + 0 * nx + 1:θ_skip + 1 * nx] .-= T0
        jac_K[m, θ_skip + 1 * nx + 1:θ_skip + 2 * nx] .+= T0
    end

end

# Parallelize Jacobian of moments
function rust_moments_trim_welfare_jacobian_threaded!(jac_K, jac_G, θ, U, obj)

    T = Threads.nthreads()
    Threads.@threads for t = 1:T
        ix0 = round(Int, (t - 1) / T * size(U)[1]) + 1
        ix1 = round(Int, t / T * size(U)[1])
        rust_moments_trim_welfare_jacobian!(@view(jac_K[ix0:ix1, :]), @view(jac_G[ix0:ix1, :, :]), θ, @view(U[ix0:ix1, :]), obj)
    end

end

# Evaluate moments and counterfactual in-place for projection inference
function rust_moments_proj!(K, G, θ, U, obj)

    G .= 0.0
    K .= 0.0

    @unpack nx, β, M00, M10, M01, M11, ccps_l, ccps_u, G_skip, θ_skip, state, σ2, nt, τ = obj.γ

    # unpack θ
    v0 = @view θ[θ_skip + 0 * nx + 1:θ_skip + 1 * nx]
    v1 = @view θ[θ_skip + 1 * nx + 1:θ_skip + 2 * nx]

    # payoffs
    π00 = payoffs_linear(@view(θ[1:2]), nx)
    π01 = payoffs_linear(@view(θ[1:2]) .* [1.0, τ], nx)

    # moments: mean-zero normalization (add any other normalizations here and adjust G_skip)
    G[:, 1:2] .= U
    G[:, 3:4] .= U.^2 .- σ2

    # moments: CCPs and value functions
    for x in 1:nx

        @views EV00 = π00[x] + β * dot(M00[x, :], v0)
        @views EV10 = β * dot(M10[x, :], v0)
        @views EV01 = π01[x] + β * dot(M01[x, :], v1)
        @views EV11 = β * dot(M11[x, :], v1)

        @inbounds for m in 1:size(U)[1]

            v00 = EV00 + U[m, 1]
            v10 = EV10 + U[m, 2]
            v01 = EV01 + U[m, 1]
            v11 = EV11 + U[m, 2]

            p0 = Φ(v10 - v00)
            p1 = Φ(v11 - v01)

            # moments for CCPs
            if x > nt
                G[m, G_skip + 0 * (nx - nt) + x - nt] = p0 - ccps_u[x]
                G[m, G_skip + 1 * (nx - nt) + x - nt] = ccps_l[x] - p0
            end

            # moments for value functions
            G[m, G_skip + 2 * nx - 2 * nt + x] = p0 * v10 + (1.0 - p0) * v00 - v0[x]
            G[m, G_skip + 3 * nx - 2 * nt + x] = p1 * v11 + (1.0 - p1) * v01 - v1[x]

            # default counterfactual is CCP in a particular state
            if x == state
                K[m] = p1
            end

        end
    end

end

# Evaluate Jacobian of moments and counterfactual in-place for projection inference
function rust_moments_proj_jacobian!(jac_K, jac_G, θ, U, obj)

    @unpack nx, β, M00, M10, M01, M11, ccps_l, ccps_u, G_skip, θ_skip, state, nt, τ = obj.γ

    jac_K .= 0.0

    # unpack θ
    v0 = @view θ[θ_skip + 0 * nx + 1:θ_skip + 1 * nx]
    v1 = @view θ[θ_skip + 1 * nx + 1:θ_skip + 2 * nx]

    # payoffs
    π00 = payoffs_linear(@view(θ[1:2]), nx)
    π01 = payoffs_linear(@view(θ[1:2]) .* [1.0, τ], nx)

    # moments: CCPs and value functions
    for x in 1:nx

        @views EV00 = π00[x] + β * dot(M00[x, :], v0)
        @views EV10 = β * dot(M10[x, :], v0)
        @views EV01 = π01[x] + β * dot(M01[x, :], v1)
        @views EV11 = β * dot(M11[x, :], v1)

        dπ0 = dpayoffs_linear(θ, x)
        dπ1 = dpayoffs_linear(θ, x) .* [1.0, τ]

        @inbounds for m in 1:size(U)[1]

            v00 = EV00 + U[m, 1]
            v10 = EV10 + U[m, 2]
            v01 = EV01 + U[m, 1]
            v11 = EV11 + U[m, 2]

            p0 = Φ(v10 - v00)
            p1 = Φ(v11 - v01)
            dp0 = dΦ(v10 - v00)
            dp1 = dΦ(v11 - v01)

            t0 = dp0 * (v10 - v00) + p0
            t1 = dp1 * (v11 - v01) + p1

            # G[m, G_skip + x - nt] = p0 - ccps_u[x]
            if x > nt
                jac_G[m, G_skip + x - nt, 1:θ_skip] .= dp0 .* -dπ0
                for y in 1:nx
                    jac_G[m, G_skip + x - nt, θ_skip + y] = dp0 * β * (M10[x, y] - M00[x, y])
                end
            end

            # G[m, G_skip + 1 * (nx - nt) + x - nt] = ccps_l[x] - p0
            if x > nt
                jac_G[m, G_skip + nx - nt + x - nt, 1:θ_skip] .= dp0 .* dπ0
                for y in 1:nx
                    jac_G[m, G_skip + nx - nt + x - nt, θ_skip + y] = -dp0 * β * (M10[x, y] - M00[x, y])
                end
            end

            # G[m, G_skip + 2 * nx - 2 * nt + x] = p0 * v10 + (1.0 - p0) * v00 - v0[x]
            jac_G[m, G_skip + 2 * nx - 2 * nt + x, 1:θ_skip] .= (1.0 - t0) .* dπ0
            jac_G[m, G_skip + 2 * nx - 2 * nt + x, θ_skip + 0 * nx + x] = 0.0
            for y in 1:nx
                if M10[x, y] > 0.0 || M00[x, y] > 0.0
                    jac_G[m, G_skip + 2 * nx - 2 * nt + x, θ_skip + 0 * nx + y] = t0 * β * (M10[x, y] - M00[x, y]) + β * M00[x, y]
                end
            end
            jac_G[m, G_skip + 2 * nx - 2 * nt + x, θ_skip + 0 * nx + x] -= 1.0

            # G[m, G_skip + 3 * nx - 2 * nt + x] = p1 * v11 + (1.0 - p1) * v01 - v1[x]
            jac_G[m, G_skip + 3 * nx - 2 * nt + x, 1:θ_skip] .= (1.0 - t1) .* dπ1
            jac_G[m, G_skip + 3 * nx - 2 * nt + x, θ_skip + 1 * nx + x] = 0.0
            for y in 1:nx
                if M11[x, y] > 0.0 || M01[x, y] > 0.0
                    jac_G[m, G_skip + 3 * nx - 2 * nt + x, θ_skip + 1 * nx + y] = t1 * β * (M11[x, y] - M01[x, y]) + β * M01[x, y]
                end
            end
            jac_G[m, G_skip + 3 * nx - 2 * nt + x, θ_skip + 1 * nx + x] -= 1.0

            # default counterfactual is CCP in a particular state
            if x == state
                jac_K[m, 1:θ_skip] .= dp1 .* -dπ1
                for y in 1:nx
                    jac_K[m, θ_skip + nx + y] = dp1 * β * (M11[x, y] - M01[x, y])
                end
            end

        end
    end

end

# Update rust_moments_trim! with new counterfactual for projection inference
function rust_moments_proj_welfare!(K, G, θ, U, obj)

    # evaluate moments as above
    rust_moments_proj!(K, G, θ, U, obj)

    # Overwrite counterfactual with welfare
    @unpack nx, θ_skip, T0 = obj.γ
    v0 = @view θ[θ_skip + 0 * nx + 1:θ_skip + 1 * nx]
    v1 = @view θ[θ_skip + 1 * nx + 1:θ_skip + 2 * nx]

    @inbounds for m in 1:size(U)[1]
        K[m] = dot(T0, v1) - dot(T0, v0)
    end

end

# Update Jacobian accordingly
function rust_moments_proj_welfare_jacobian!(jac_K, jac_G, θ, U, obj)

    # Evaluate Jacobian as above
    rust_moments_proj_jacobian!(jac_K, jac_G, θ, U, obj)

    # Overwrite counterfactual Jacobian with welfare
    @unpack nx, θ_skip, T0 = obj.γ

    @inbounds for m in 1:size(U)[1]
        jac_K[m, θ_skip + 0 * nx + 1:θ_skip + 1 * nx] .-= T0
        jac_K[m, θ_skip + 1 * nx + 1:θ_skip + 2 * nx] .+= T0
    end

end

# Parallelize Jacobian of moments
function rust_moments_proj_welfare_jacobian_threaded!(jac_K, jac_G, θ, U, obj)

    T = Threads.nthreads()
    Threads.@threads for t = 1:T
        ix0 = round(Int, (t - 1) / T * size(U)[1]) + 1
        ix1 = round(Int, t / T * size(U)[1])
        rust_moments_proj_welfare_jacobian!(@view(jac_K[ix0:ix1, :]), @view(jac_G[ix0:ix1, :, :]), θ, @view(U[ix0:ix1, :]), obj)
    end

end
