# Codes to plot all figures for the Rust example

# Assumes results have been generated by running
# rust_main.jl,
# rust_proj.jl,
# rust_phi.jl

# Load packages, etc
import Pkg; Pkg.activate("../..")
using KNITRO, ForwardDiff, Calculus, LinearAlgebra, Distributions, Parameters, Random, JLD, RData, Optim, Plots, LaTeXStrings, BenchmarkTools, ColorSchemes, StatsBase
include("../../src/boot.jl")        # Load main programs
include("./aux/rust_boot.jl")       # Load model-specific files and programs

# Level of significance
α = 0.05

# Save figures?
saving = true

# Formatting
labels = ("δ = 0.01", "δ = 0.1", "δ = 1", "δ = 10", "δ = 100")
colors = [palette(:cubehelix, 15)[i] for i in (1, 3, 5, 8, 9, 11)]

# Load saved objects from rust_initialize.jld
x = load("rust_initialize.jld")
P = x["P"]
β = x["β"]
τ = x["τ"]
κ_boot = x["κ_boot"]
κ0 = x["κ0"]
nx = x["nx"]
θπ = x["θπ"]
T0 = x["T0"]
M00 = x["M00"]
M01 = x["M01"]
M10 = x["M10"]
M11 = x["M11"]

# Trim CCPs less than 0.001;
trim = 0.001
nt = length(findall(P .< trim))

# 4 location/scale normalizations
G_skip = 4

# ObjectiveBundle for welfare counterfactual
obj = PsiObjectiveBundleImplicit(
    δ = 0.01,
    find_smallest = true,
    γ = (nx = nx, β = β, M00 = M00, M10 = M10, M01 = M01, M11 = M11, ccps = P, θ_skip = 2, G_skip = G_skip, state = -1, T0 = T0, σ2 = Base.MathConstants.pi^2/6.0, nt = nt, τ = τ),
    moments! = rust_moments_trim_welfare!,
    moments_jacobian! = rust_moments_trim_welfare_jacobian_threaded!,
    d = 2 * nx + (nx - nt) + G_skip,
    outer_constr_index = (nx - nt) + G_skip + 1,
    inequality_index = Int64[],
    l = 2 + 2 * nx,
    U = quantile.(GeneralizedExtremeValue(-Base.MathConstants.eulergamma, 1, 0), rhalton(50000, 2, singleseed = 1234567)),
    outer_loop_opt = "./opt/rust_outer_loop_settings_delta.opt",
    inner_loop_opt = "./opt/rust_inner_loop_settings_implicit.opt",
    lower_limit = -1e3,
    use_cached_x = false)

# Part 1: Plot main estimates and CIs
δ_grid = 10.0.^vcat(-2.0:(1/2):2.0)
δ_refs = (1:2:9)
δ_plot = δ_grid[δ_refs]

κ = zeros(length(δ_plot), 3, 2)

# Trim bootstrap perturbations and adjust for location/scale normalizations
B = length(κ_boot)
z = zeros(nx - nt + G_skip, B)
z[G_skip + 1:end, :] .= x["z"][nt + 1:end, :]

# Compute bootstrap CIs
for array_value in 1:2

    # Load counterfactual estimates (generated by rust_main.jl)
    local x = load(string("./results/rust_main_",array_value,".jld"))
    κ[:, 1, array_value] .= x["κ"][δ_refs]

    # Compute bootstrap critical value with nu hat = 0
    for (i, j) in enumerate(δ_refs)

        # Using outer_loop_bootstrap
        # c = zeros(B)
        # for b in 1:B
        #     c[b] = outer_loop_bootstrap(obj, x["Θ"][:, j], z[:, b], x["Λ"][:, j])
        # end

        # Numerically equivalent to outer_loop_bootstrap but avoids repeated computation of optimization problem
        inner_loop(obj, x["Θ"][:, j])
        obj.x .*= x["Λ"][:, j][1] * 1e10
        c = BLAS.gemv('T', 1.0, z, obj.x[2:end])

        if array_value == 1
            κ[i, 2, 1] = κ[i, 1, 1] - quantile(c, α)
        else
            κ[i, 2, 2] = κ[i, 1, 2] - quantile(c, 1 - α)
        end
    end

end

# Plot
xgrid = log.((δ_plot)) ./ log(10)
p1 = plot()
plot!(p1, xgrid, κ[:, 1, 1], lw = 3, ms = 4, ls = :solid, markershape = :square, color = colors[3], label = "")
plot!(p1, xgrid, κ[:, 1, 2], lw = 3, ms = 4, ls = :solid, markershape = :square, color = colors[3], label = "")
plot!(p1, xgrid, κ[:, 1, 1], lw = 3, ls = :solid, color = colors[3], label = "Bounds")
plot!(p1, xgrid, κ[:, 2, 1], lw = 3, ms = 4, ls = :dot, color = colors[3], label = "")
plot!(p1, xgrid, κ[:, 2, 2], lw = 3, ms = 4, ls = :dot, color = colors[3], label = "")
plot!(p1, xgrid, κ0 * ones(length(δ_plot)), lw = 3, color = :black, label = "Gumbel")
plot!(p1, xgrid, (κ0 - quantile(κ_boot .- κ0, α)) * ones(length(δ_plot)), lw = 3, color = :black, ls = :dot, label = "")
plot!(p1, xgrid, (κ0 - quantile(κ_boot .- κ0, 1 - α)) * ones(length(δ_plot)), lw = 3, color = :black, ls = :dot, label = "")
plot!(p1, xticks = (xgrid, ["0.01", "0.1", "1", "10", "100"]))
xlabel!(p1, "δ (log scale)", xlabelfontsize = 12)
ylabel!(p1, "Change in Avgerage Welfare", ylabelfontsize = 12)
plot!(p1, ylims = (-5.0, 205.0))
plot!(p1, xtickfontsize = 12, ytickfontsize = 12)
plot!(p1, legendfontsize = 12, legend = :topleft)
saving ? savefig("rust_bounds.pdf") : nothing

# Part 2: Plot LFDs
Ρ = zeros(length(δ_plot), 4, 2)
for array_value in 1:2

    local p0 = plot(x -> exp(-exp(-x-Base.MathConstants.eulergamma)), ls = :dot, lw = 5, color = colors[1], label = "δ = 0")
    local p1 = plot(x -> exp(-exp(-x-Base.MathConstants.eulergamma)), ls = :dot, lw = 5, color = colors[1], label = "")
    local p2 = plot(x -> exp(x)/(1+exp(x)), ls = :dot, lw = 5, color = colors[1], label = "δ = 0")

    for (i,j) in enumerate(δ_refs)
        local x = load(string("./results/rust_main_",array_value,".jld"))
        inner_loop(obj, x["Θ"][:, j])
        i > 1 ? plot!(p0, quantile(obj.U[:, 1], Weights(obj.arg1), 0.0:0.01:1.0), 0.0:0.01:1.0, lw = 4, color = colors[i + 1], label = labels[i]) : nothing
        i > 1 ? plot!(p1, quantile(obj.U[:, 2], Weights(obj.arg1), 0.0:0.01:1.0), 0.0:0.01:1.0, lw = 4, color = colors[i + 1], label = "") : nothing
        plot!(p2, quantile(obj.U[:, 2] - obj.U[:, 1], Weights(obj.arg1), 0.0:0.01:1.0), 0.0:0.01:1.0, lw = 4, color = colors[i + 1], label = labels[i])

        # Correlation under LFD
        Ρ[i, 1, array_value] = sum(obj.U[:, 1] .* obj.U[:, 2] .* obj.arg1) / (obj.M * obj.γ.σ2)
        # Size measure
        Ρ[i, 2, array_value] = maximum(abs.([mean(obj.H[:, x]) for x in 7:77]))
        # Cost parameters
        Ρ[i, 3:4, array_value] .= x["Θ"][1:2, j]

    end
    plot!(p0, xlims = (-2.2, 3.2), xtickfontsize = 16, ytickfontsize = 16)
    plot!(p1, xlims = (-2.2, 3.2), xtickfontsize = 16, ytickfontsize = 16)
    plot!(p2, xlims = (-4, 4), xtickfontsize = 16, ytickfontsize = 16)
    plot!(p0, legend = :bottomright, legendfontsize = 16)
    plot!(p2, legend = array_value == 2 ? :bottomright : nothing, legendfontsize = 16)
    if saving
        savefig(p0, string("rust_lfd_",array_value,"_0.pdf"))
        savefig(p1, string("rust_lfd_",array_value,"_1.pdf"))
        savefig(p2, string("rust_lfd_",array_value,".pdf"))
    end

end

# Tabulate
Ρ[:, 4, :] .*= -1e3
Ρ .= round.(Ρ, digits = 3)
println()
for (i, δ) in enumerate(δ_plot)
    println(string(δ, " & & ", Ρ[i, 1, 2], " & ", Ρ[i, 2, 2], " & ", Ρ[i, 3, 2], " & ", Ρ[i, 4, 2], " & & ", Ρ[i, 1, 1], " & ", Ρ[i, 2, 1], " & ", Ρ[i, 3, 1], " & ", Ρ[i, 4, 1], " \\\\"))
end
println()

# Part 3: Computation times
for array_value in 1:2
    for (i, j) in enumerate(δ_refs)
        local x = load(string("./results/rust_main_",array_value,".jld"))
        inner_loop(obj, x["Θ"][:, j])
        println([array_value, δ_grid[j]])
        @btime inner_loop_KNITRO(obj)
        println()
    end
    println()
end

# Part 4: Projection CSs
for array_value in 1:2

    # Load counterfactual estimates (generated by rust_main.jl)
    local x = load(string("./results/rust_proj_",array_value,".jld"))
    κ[:, 3, array_value] .= x["κ"][δ_refs]

end

# Plot
plot!(p1, xgrid, κ[:, 3, 1], lw = 3, ms = 4, ls = :dash, color = colors[3], label = "")
plot!(p1, xgrid, κ[:, 3, 2], lw = 3, ms = 4, ls = :dash, color = colors[3], label = "")
plot!(p1, ylims = (-5.0, 255.0))
plot!(p1, xtickfontsize = 12, ytickfontsize = 12)
plot!(p1, legendfontsize = 12, legend = :topleft)
saving ? savefig("rust_bounds_2.pdf") : nothing

# Part 5: Summarize results with different phi functions
κ_phi = zeros(length(δ_plot), 3, 2)
κ_phi[:, 1, :] .= κ[:, 1, :]
κ_phi[:, 2, 1] .= load("./results/rust_phi_1.jld")["κ"][δ_refs]
κ_phi[:, 3, 1] .= load("./results/rust_phi_2.jld")["κ"][δ_refs]
κ_phi[:, 2, 2] .= load("./results/rust_phi_3.jld")["κ"][δ_refs]
κ_phi[:, 3, 2] .= load("./results/rust_phi_4.jld")["κ"][δ_refs]

println()
println("width of χ² bounds relative to hybrid bounds")
println(round.((κ_phi[:, 2, 1] .- κ_phi[:, 2, 2])./(κ_phi[:, 1, 1] .- κ_phi[:, 1, 2]), digits = 2))
println()
println("width of L⁴ bounds relative to hybrid bounds")
println(round.((κ_phi[:, 3, 1] .- κ_phi[:, 3, 2])./(κ_phi[:, 1, 1] .- κ_phi[:, 1, 2]), digits = 2))
println()
