# Codes to initialize Rust example

# Load packages, etc
import Pkg; Pkg.activate("../..")
using KNITRO, ForwardDiff, Calculus, LinearAlgebra, Distributions, Parameters, Random, JLD, RData, Optim
include("../../src/boot.jl")        # Load main programs
include("./aux/rust_boot.jl")       # Load model-specific files and programs

# Parameters
β = 0.9999              # Discount factor
nx = 90                 # Dimension of state-space as in Rust (1987), Table IX
B = 1000                # Number of bootstrap draws
τ = 0.9                 # Reduction in MC parameter in counterfactual

# Load Rust data
data = load("./data/a530875.RData")
data["dtc"] = convert(Array{Int64,2}, data["dtc"])
data["dtx"] = convert(Array{Int64,2}, data["dtx"])
data["mil"] = convert(Array{Int64,2}, data["mil"])

# Set up transition matrices
M00, M10 = make_transition_matrices(nx, data["mil"])
M01, M11 = copy(M00), copy(M10)

# Compute MLE for θπ
(θπ, H) = rust_mle(data, nx, β, M00, M10)

# Benchmark payoffs as estimated
π0 = payoffs_linear(θπ, nx)
v0 = fixed_point(π0, β, M00, M10)

# CCPs at MLE
P = ccps_logit(π0, β, M00, M10, v0)

# Stationary distrbution of the state
T = M10 .* P + M00 .* (1.0 .- P)
T0 = (T^1000)[1, :]

# Counterfactual payoffs
π1 = payoffs_linear(θπ .* [1.0, τ], nx)
v1 = fixed_point(π1, β, M01, M11)

# Initial parameter and counterfactual
θ0 = vcat(θπ, v0, v1)
κ0 = dot(T0, v1 - v0)

# Perturbations for the bootstrap
Random.seed!(1234567)
z = zeros(nx, B)
κ_boot = zeros(B)
θ_boot = zeros(length(θ0), B)
P_boot = zeros(nx, B)
for b in 1:B
    θπ_boot = rand(MvNormal(θπ, 0.5 * (inv(H) + inv(H)')))
    π0_boot = payoffs_linear(θπ_boot, nx)
    v0_boot = fixed_point(π0_boot, β, M00, M10)
    π1_boot = payoffs_linear(θπ_boot .* [1.0, τ], nx)
    v1_boot = fixed_point(π1_boot, β, M01, M11)
    ccps_boot = ccps_logit(π0_boot, β, M00, M10, v0_boot)
    global z[:, b] .= ccps_boot .- P
    global P_boot[:, b] .= ccps_boot
    global θ_boot[:, b] .= vcat(θπ_boot, v0_boot, v1_boot)
    global κ_boot[b] = dot(T0, v1_boot - v0_boot)
end

save("rust_initialize.jld", "τ", τ, "θ0", θ0, "κ0", κ0, "κ_boot", κ_boot, "P_boot", P_boot, "θ_boot", θ_boot, "z", z, "P", P, "θπ", θπ, "T0", T0, "nx", nx, "β", β, "M00", M00, "M10", M10, "M01", M01, "M11", M11)
