# Codes to generate estimates for the Rust (1987) example with different ϕ-divergences
# Run rust_initialize.jl first to generate transition matrices, CCP estimates, etc

# Load packages, etc
import Pkg; Pkg.activate("../..")
using KNITRO, ForwardDiff, Calculus, LinearAlgebra, Distributions, Parameters, Random, JLD, RData, Optim
include("../../src/boot.jl")        # Load main programs
include("./aux/rust_boot.jl")       # Load model-specific files and programs

array_value = Base.parse(Int, ENV["SLURM_ARRAY_TASK_ID"])
finding_smallest = array_value <= 2 ? false : true
phi_choice = array_value ∈ (1, 3) ? 1 : 2
println([Threads.nthreads(), array_value, finding_smallest]); flush(stdout)

# Load saved objects from rust_initialize.jld
x = load("rust_initialize.jld")
τ = x["τ"]      # subsidy parameter
P = x["P"]      # estimated CCPs
β = x["β"]      # discount factor
nx = x["nx"]    # number of states
θ1 = x["θ0"]    # parameter at estimated CCPs
T0 = x["T0"]    # stationary distribution of state
M00 = x["M00"]  # pre-intervention M0 transition matrix
M01 = x["M01"]  # pre-intervention M1 transition matrix
M10 = x["M10"]  # post-intervention M0 transition matrix
M11 = x["M11"]  # post-intervention M1 transition matrix

# Trim CCPs less than 0.001;
trim = 0.001
nt = length(findall(P .< trim))

# 4 location/scale normalizations
G_skip = 4

# Lower and upper bounds on parameter space for θ
θ_lower = vcat(0.0, -10.0, zeros(2 * nx))
θ_upper = vcat(20.0, 0.0, 1e6*ones(2 * nx))

# ObjectiveBundle for welfare counterfactual
obj = PsiObjectiveBundleImplicit(
    δ = 0.01,
    find_smallest = finding_smallest,
    γ = (nx = nx, β = β, M00 = M00, M10 = M10, M01 = M01, M11 = M11, ccps = P, θ_skip = 2, G_skip = G_skip, state = -1, T0 = T0, σ2 = Base.MathConstants.pi^2/6.0, nt = nt, τ = τ),
    moments! = rust_moments_trim_welfare!,
    moments_jacobian! = rust_moments_trim_welfare_jacobian_threaded!,
    d = 2 * nx + (nx - nt) + G_skip,
    outer_constr_index = (nx - nt) + G_skip + 1,
    inequality_index = Int64[],
    l = 2 + 2 * nx,
    U = quantile.(GeneralizedExtremeValue(-Base.MathConstants.eulergamma, 1, 0), rhalton(50000, 2, singleseed = 1234567)),
    outer_loop_opt = "./opt/rust_outer_loop_settings_implicit.opt",
    inner_loop_opt = "./opt/rust_inner_loop_settings_implicit.opt",
    lower_limit = -1e3,
    use_cached_x = false)

# Pre-allocate to store estimates for outer optimization across different δ
δ_grid = 10.0.^vcat(-2.0:(1/2):2.0)
κ = zeros(length(δ_grid))
flag = zeros(length(δ_grid))
Θ = zeros(obj.l, length(δ_grid))

# Estimate
if phi_choice == 1
    obj.Psi! = Psi_χ2!
    obj.dPsi! = dPsi_χ2!
    obj.ddPsi! = ddPsi_χ2!
elseif phi_choice == 2
    obj.Psi! = Psi_l4!
    obj.dPsi! = dPsi_l4!
    obj.ddPsi! = ddPsi_l4!
end

for (i, δ) in enumerate(δ_grid)
    println(δ); flush(stdout)
    obj.δ = δ
    obj.lower_limit = -1.1 * δ
    (κ_temp, θ1_temp, flag_temp) = outer_loop(obj, θ_lower, θ_upper, θ1)
    if i == 1 || (obj.find_smallest && κ_temp < κ[i-1]) || (!obj.find_smallest && κ_temp > κ[i-1])
        global (κ[i], θ1, flag[i]) = (κ_temp, θ1_temp, flag_temp)
        global Θ[:, i] .= θ1
    else
        global (κ[i], θ1, flag[i]) = (κ[i-1], Θ[:, i-1], flag[i-1])
        global Θ[:, i] .= θ1
    end
    save(string("./results/rust_phi_",array_value,".jld"),"δ_grid", δ_grid, "κ", κ, "Θ", Θ, "flag", flag)
    if flag[i] < -103.0
        println([δ, flag[i]]); flush(stdout)
        (κ_temp, θ1_temp, flag_temp) = outer_loop(obj, θ_lower, θ_upper, θ1)
        if flag_temp >= -103.0 || ((obj.find_smallest && κ_temp < κ[i]) || (!obj.find_smallest && κ_temp > κ[i]))
            global (κ[i], θ1, flag[i]) = (κ_temp, θ1_temp, flag_temp)
            global Θ[:, i] .= θ1
        end
        save(string("./results/rust_phi_",array_value,".jld"),"δ_grid", δ_grid, "κ", κ, "Θ", Θ, "flag", flag)
    end
end
