# Rust (1987) example -- timing for the Norets-Tang design

# Load packages, etc
import Pkg; Pkg.activate("../..")
using KNITRO, ForwardDiff, Calculus, LinearAlgebra, Distributions, Parameters, Random, JLD, RData, BenchmarkTools
include("../../src/boot.jl")        # Load main programs
include("./aux/rust_boot.jl")       # Load model-specific files and programs

# Fixed parameters
β = 0.999               # Discount factor
nx = 90                 # Dimension of state-space as in Rust (1987), Table IX

# Set up transition matrices
M10 = zeros(nx, nx)
M10[:, 1] .= 1.0
M11 = copy(M10)
M00 = transition_matrix(nx, [0.3919, 0.5953])
M01 = transition_matrix(nx, [0.6, 0.3])

# Payoffs and value functions`
θπ = [5.0727, -0.002293]
π0 = payoffs_linear(θπ, nx)
v0 = fixed_point(π0, β, M00, M10)
π1 = copy(π0)
v1 = fixed_point(π1, β, M01, M11)
θ_initial = vcat(θπ, v0, v1)

# CCPs`
P = ccps_logit(π0, β, M00, M10, v0)

G_skip = 4
δ_grid = 10.0.^vcat(-2.0:1.0:0.0)

# MPEC implementation
obj = PsiObjectiveBundleExplicit(
    δ = 0.01,
    find_smallest = false,
    γ = (nx = nx, β = β, M00 = M00, M10 = M10, M01 = M01, M11 = M11, ccps = P, θ_skip = 2, G_skip = G_skip, state = 90, trim = 0.0, σ2 = Base.MathConstants.pi^2/6.0, τ = 1.0),
    moments! = rust_moments!,
    moments_jacobian! = rust_moments_jacobian_threaded!,
    d = 3 * nx + G_skip,
    outer_constr_index = nx + G_skip + 1,
    inequality_index = [],
    l = 2 + 2 * nx,
    U = quantile.(GeneralizedExtremeValue(-Base.MathConstants.eulergamma, 1, 0), rhalton(50000, 2, singleseed = 1234567)),
    outer_loop_opt = "./opt/rust_outer_loop_settings_implicit.opt",
    inner_loop_opt = "./opt/rust_inner_loop_settings_implicit.opt",
    lower_limit = -1e3,
    use_cached_x = false)

println("Counterfactual CCP, MPEC")
for δ in δ_grid
    println(δ)
    obj.δ = δ
    inner_loop(obj, θ_initial)
    @btime inner_loop_KNITRO(obj)
end
println()

# non-MPEC implementation
obj = PsiObjectiveBundleExplicit(
    δ = 0.01,
    find_smallest = false,
    γ = (nx = nx, β = β, M00 = M00, M10 = M10, M01 = M01, M11 = M11, ccps = P, θ_skip = 2, G_skip = G_skip, state = 90, trim = 0.0, σ2 = Base.MathConstants.pi^2/6.0, τ = 1.0),
    moments! = rust_moments!,
    moments_jacobian! = rust_moments_jacobian_threaded!,
    d = 3 * nx + G_skip,
    inequality_index = [],
    l = 2 + 2 * nx,
    U = quantile.(GeneralizedExtremeValue(-Base.MathConstants.eulergamma, 1, 0), rhalton(50000, 2, singleseed = 1234567)),
    outer_loop_opt = "./opt/rust_outer_loop_settings_implicit.opt",
    inner_loop_opt = "./opt/rust_inner_loop_settings_implicit.opt",
    lower_limit = -1e3,
    use_cached_x = false)

println("Counterfactual CCP, non-MPEC")
for δ in δ_grid
    println(δ)
    obj.δ = δ
    inner_loop(obj, θ_initial)
    @btime inner_loop_KNITRO(obj)
end
println()

# MPEC implementation
obj = PsiObjectiveBundleDelta(
    γ = (nx = nx, β = β, M00 = M00, M10 = M10, M01 = M01, M11 = M11, ccps = P, θ_skip = 2, G_skip = G_skip, state = 90, trim = 0.0, σ2 = Base.MathConstants.pi^2/6.0, τ = 1.0),
    moments! = rust_moments!,
    moments_jacobian! = rust_moments_jacobian_threaded!,
    d = 3 * nx + G_skip,
    outer_constr_index = nx + G_skip + 1,
    inequality_index = [],
    l = 2 + 2 * nx,
    U = quantile.(GeneralizedExtremeValue(-Base.MathConstants.eulergamma, 1, 0), rhalton(50000, 2, singleseed = 1234567)),
    outer_loop_opt = "./opt/rust_outer_loop_settings_implicit.opt",
    inner_loop_opt = "./opt/rust_inner_loop_settings_implicit.opt",
    lower_limit = -1e3,
    use_cached_x = false)

println("Minimum Divergence, MPEC")
inner_loop(obj, θ_initial)
@btime inner_loop_KNITRO(obj)
println()

# non-MPEC implementation
obj = PsiObjectiveBundleDelta(
    γ = (nx = nx, β = β, M00 = M00, M10 = M10, M01 = M01, M11 = M11, ccps = P, θ_skip = 2, G_skip = G_skip, state = 90, trim = 0.0, σ2 = Base.MathConstants.pi^2/6.0, τ = 1.0),
    moments! = rust_moments!,
    moments_jacobian! = rust_moments_jacobian_threaded!,
    d = 3 * nx + G_skip,
    inequality_index = [],
    l = 2 + 2 * nx,
    U = quantile.(GeneralizedExtremeValue(-Base.MathConstants.eulergamma, 1, 0), rhalton(50000, 2, singleseed = 1234567)),
    outer_loop_opt = "./opt/rust_outer_loop_settings_implicit.opt",
    inner_loop_opt = "./opt/rust_inner_loop_settings_implicit.opt",
    lower_limit = -1e3,
    use_cached_x = false)

println("Minimum Divergence, non-MPEC")
inner_loop(obj, θ_initial)
@btime inner_loop_KNITRO(obj)
println()
