# Generate figures for toy example (Example 2.4)
using Optim, Distributions, ForwardDiff, Plots, LaTeXStrings

Φ(x) = Distributions.cdf(Normal(0, 1), x)
φ(x) = Distributions.logcdf(Normal(0, 1), x)

# Example: max Pr(U ≦ θ) subject to mean-θ restriction
function f1(η, λ, δ, θ)

    # f = λ^2 / (2.0 * η) + η * log(1.0 + (exp(1.0 / η) - 1.0) * Φ(λ / η)) + η * δ
    # use numerically sound log-sum-exp to avoid underflow
    t1 = φ(-θ -λ / η)
    t2 = 1.0 / η + φ(θ + λ / η)
    tm = max(t1, t2)
    f = λ * θ + λ^2 / (2.0 * η) + η * δ + η * tm + η * log(exp(t1 - tm) + exp(t2 - tm))

    return f

end

function v(δ, ff, θ)

    function f(x)
        return ff(x[1], x[2], δ, θ)
    end

    function g!(storage, x)
        storage .= ForwardDiff.gradient(f, x)
    end

    lower = [1e-120, -Inf]
    upper = [Inf, Inf]

    results = optimize(f, g!, lower, upper, [1.0, 0.0], Fminbox(GradientDescent()), Optim.Options(g_tol = 1e-16, iterations = 10000))

    x = Optim.minimizer(results)
    κ = Optim.minimum(results)
    flag = Optim.converged(results)

    return (x, κ, flag)

end

# Illustrate asymptotics are correct
for θ in (0.0, 0.8)
    δ_grid = 0.5:0.5:5.0
    η_grid = zeros(size(δ_grid))
    λ_grid = zeros(size(δ_grid))
    κ_grid = zeros(size(δ_grid))
    f_grid = zeros(size(δ_grid))
    for (i, δ) in enumerate(δ_grid)
        ((η_grid[i], λ_grid[i]), κ_grid[i], f_grid[i]) = v(δ, f1, θ)
    end

    # Illustrate asymptotic result for η
    plot(δ_grid, log.(η_grid), lw = 2, label = "Actual", c = :black)
    plot!(δ_grid, log(4 * π) .- (1.0 - θ^2) .- 2.0 .* δ_grid, lw = 2, label = "Asymptotic", c = :gray)
    plot!(xlabel = L"\delta", title = L"\theta = %$θ", thickness_scaling = 4/3)
    θ == 0.0 ? plot!(ylabel = L"\log(\eta)") : plot!(legend = false)
    savefig(string("eta_plot_",θ,".pdf"))
end


# Plot objective function and approximation for different δ
θ_grid = -1.0:0.05:1.0
K_grid = zeros(size(θ_grid))
for δ in 1.0:4.0
    for (i, θ) in enumerate(θ_grid)
        ((η, λ), K_grid[i], flag) = v(δ, f1, θ)
        plot(θ_grid, K_grid, lw = 2, label = "Actual", c = :black)
        plot!(θ_grid, 1.0 .- 2 * π * exp.(.-(1.0 .- θ_grid .^ 2) .- 2.0 .* δ), lw = 2, label = "Asymptotic", c = :gray)
        Δ = Int(δ)
        plot!(title = L"\delta = %$Δ")
        δ == 1.0 ? plot!(legend = :bottom) : plot!(legend = false)
        δ ∈ (3.0, 4.0) ? plot!(xlabel = L"\theta") : nothing
        δ ∈ (1.0, 3.0) ? plot!(ylabel = L"\overline{K}\!_\delta(\theta)") : nothing
        plot!(thickness_scaling = 4/3)
        savefig(string("K_plot_",Δ,".pdf"))
    end
end
