# Guide to General-Purpose Codes for Christensen and Connault (2022)

**Disclaimer:** This material is based upon work supported by the **National Science Foundation** under Grant Number **SES-1919034**. Any opinions, findings, and conclusions or recommendations expressed in this material are those of the author(s) and do not necessarily reflect the views of the National Science Foundation.

## Software requirements

You will need:

* Julia version 1.6.1 +
* Artelys Knitro version 12.4.0 +

All scripts have been tested in Julia v1.6.4 and Knitro v12.4.0 on macOS X. The code in this repository depends on other packages. Run `import Pkg, Random; Pkg.activate("."); Pkg.instantiate()` in the root directory to install all necessary packages.

## Basics

The main functions are `inner_loop` and `outer_loop`, which perform the inner optimization with respect to Lagrange multipliers and the outer optimization with respect to structural parameters, respectively.

The key input is an `ObjectiveBundle`. This object collects all the information about the model and optimization problems. We outline its construction below.

The **inner** optimization at a particular value of `θ` is executed using the command
```
(κ, x, flag) = inner_loop(obj, θ)
```
where `obj` is the `ObjectiveBundle`. The outputs are
* `κ` the smallest or largest counterfactual at fixed `θ`
* `x` a vector collecting the Lagrange multipliers at which the optimal value is attained
* `flag` a Knitro solution status return code.

The **outer** optimization with respect to θ is executed using the command
```
(κ, θ, flag) = outer_loop(obj, θ_lower, θ_upper, θ_initial)
```
where
* `θ_lower` and `θ_upper` are element-wise bounds on θ
* `θ_initial` is an initial point at which to begin the optimization.

The outputs are
* `κ` the smallest or largest counterfactual
* `θ` the vector of parameters at which the optimal value is attained
* `flag` a Knitro solution status return code.

Calling `(κ, θ, flag, Λ) = outer_loop(...)` will also return a vector `Λ` of multipliers on the constraints in the outer loop. These are, in order, (i) multiplier for the neighborhood constraint 1e10 * (Δ(θ; γ, P) - δ) <= 0 if an implicit counterfactual, (ii) MPEC constraints, if any, then (iii) multipliers for the element-wise bounds on θ.

## User inputs

The prefix of `ObjectiveBundle` should be `KL` for neighborhoods constrained by Kullback-Leibler divergence or `Psi` for neighborhoods constrained by any user-specified phi-divergence. Its suffix should be:
* `Explicit` for counterfactual functions that depend on both structural parameters and unobservables
* `Implicit` for counterfactual functions that depend only on structural parameters
* `Delta` for the minimum divergence problem

**Example:** use `KLObjectiveBundleExplicit` for KL neighborhoods and a counterfactual function that depends on structural parameters and unobservables.

**Example:** use `PsiObjectiveBundleDelta` for the minimum-divergence problem using hybrid divergence (by default) or other phi-divergence (see below) neighborhoods.

The minimal set of inputs to be provided when defining the `ObjectiveBundle` are:

* `U` a matrix of `M` rows collecting the random draws of unobservables from their benchmark distribution to be used for Monte Carlo integration
* `moments!` a function which evaluates the moment functions and counterfactual functions in-place across each of the `M` draws, which should be called as
```
moments!(K, G, θ, U, obj)
```
where
  * `K` is `M x 1` and collects the counterfactual function evaluated at each of the `M` draws (fill all entries to zeros for the minimum divergence problem)
  * `G` is `M x d` and collects the moment functions evaluated at each of the `M` draws
  * `θ` is the vector of structural parameters
  *  `obj` is the objective bundle
* `d` the number of columns in `G`
* `l` the length of `θ`
* `inequality_index` a vector listing the columns of `G` that correspond to moment inequalities; set `inequality_index = Int64[]` if all are equalities
* `γ` a tuple collecting any additional objects required for evaluating the moments
* `inner_loop_opt` Knitro options file for the inner optimization (see examples for default settings)
* `outer_loop_opt` Knitro options file for the outer optimization (see examples for default settings)

In addition, for optimizing counterfactuals (either the `Explicit` or `Implicit` case) the user should provide
* `δ` the neighborhood size
* `find_smallest = true` to find the smallest  value of the counterfactual or `= false` to find the largest value

**Note:** reduced-form parameters should be subtracted from the moment function in evaluation of `moments!` so that the expected value of the moments are zero (if equality constraints) or non-positive (if inequality constraints) when satisfied. Any reduced-form parameters can be stored in `γ` and passed to `moments!` through `obj`.

## Performance tips

* Set `gradopt = exact` in `outer_loop_opt` to use explicit gradients and Jacobians in the outer optimization.

  By default, gradients and Jacobians will be calculated by automatic differentiation across all `M` draws. For this to work, moments should be smooth, which may require smoothing out indicator functions, maxima/minima, etc.

  Default Jacobians are computed for each draw. In models where the number of moments and/or parameters is large, storing the whole Jacobian may be infeasible due to memory constraints. The user can specify a value for `N` in the `ObjectiveBundle` to compute approximate gradients and Jacobians using only the first `N` draws, saving memory. The parameter `N` should be set when defining the `ObjectiveBundle`, as it defaults to `M`

  Analytic gradients can instead be provided directly through specifying `moments_jacobian!` in the `ObjectiveBundle`. This will override autodiff. Again, these can be provided using a smaller number of `N` draws, saving memory.

* Set `convex = 1` in `inner_loop_opt` to declare the inner problem as convex.
* The user should experiment to make evaluation of `moments!` (and `moments_jacobian!`, if using) as efficient as possible.
* If performing projection inference, be sure to impose complementarity constraints for Lagrange multipliers corresponding to the upper and lower CI for each moment (see below).
* The function `outer_loop_multi` implements a multi-start version of `outer_loop` with a more selective handling of return codes than Knitro.

## MPEC implementation

To activate, order the columns of `G` so that columns `1` to  `d0` collect the moments used in the inner optimization and columns `d0 + 1` to `d` collect the moments to be used as equilibrium constraints in the outer optimization. Then set `outer_constr_index = d0 + 1` as an option in `ObjectiveBundle`.

**Example:** suppose the model has 20 moments, of which 6 are equilibrium conditions to be used as constraints in the outer optimization. The user would write `moments!` so that `G` stores the 14 moments for the inner optimization in columns 1 to 14 and the second set of 6 moments in columns 15 to 20, and set `outer_constr_index = 15`.

Omitting `outer_constr_index` or setting `outer_constr_index = d + 1` where `d` is the number of columns in `G` will de-active the MPEC implementation.

## Other divergences

Any user-supplied phi-divergence can be used directly by defining a `PsiObjectiveBundleExplicit`, `PsiObjectiveBundleImplicit`, or `PsiObjectiveBundleDelta` and editing the functions `Psi!`, `dPsi!`, and `ddPsi!` in `Psi.jl`, which evaluate the convex conjugate of phi and its first and second derivatives in-place.

## Bootstrap inference

Performing bootstrap inference requires repeatedly computing a min-max or max-min statistic across a large number of draws of random vectors approximating the distribution of $n^{1/2}(\hat P - P_0)$. For a `PsiObjectiveBundle` of type `Explicit` or `Delta`, the statistic is computed for a given draw `Z` using the command

```
(c, θ, flag) = outer_loop_bootstrap(obj, θ_lower, θ_upper, θ_hat, Z, ε)
```
where
* `obj`, `θ_lower`, and `θ_upper` are as described above
* `θ_hat` is the minimizing or maximizing value of θ computed using `outer_loop`
* `ε` is a threshold that should go to zero with sample size n but at a rate slower than $n^{1/2}$.

The outputs are
* `c` the min-max or max-min statistic
* `θ` the vector of parameters at which the optimal value is attained
* `flag` a Knitro solution status return code.

For a `PsiObjectiveBundleImplicit`, the statistic is computed using
```
c = outer_loop_bootstrap(obj, θ_hat, Z, Λ)
```
where `obj` and `θ_hat` are as described above and `Λ` is returned by `outer_loop` or `outer_loop_multi` when computing `θ_hat`. This approach approximates the min-max and max-min statistics at `θ_hat` rather than optimizing over an estimate of the minimizing or maximizing set of θ values. It is therefore valid, but possibly conservative.

**Note:** `outer_loop_bootstrap` is configured for models defined by moment equalities and for which all moment functions are linearly independent.

Refer to the replication files for the Chiappori et al. (2017) and Rust (1987) examples to see how to implement this command in practice.


## Projection inference

Projection inference is performed using `outer_loop`, modifying `moments!` (and `moments_jacobian!`, if using) with moment inequalities formed using a rectangular confidence set for the reduced-form parameters.

**Note:** Moment equalities of the form $E[g(U,θ)] = P$ are converted to two inequality restrictions $E[g(U,θ)] - P_U ≤ 0$ and $P_L - E[g(U,θ)] ≤ 0$ formed from the confidence interval $[P_L, P_U]$ for P. The Lagrange multipliers corresponding to the moment inequalities formed from the upper and lower CI bounds are complementary: at most one is binding.

If the model features complementary moments, pass `complement_index` (a `j x 2` matrix of integers) as an input to `obj`. The first column of `complement_index` should list which of the `j` columns of `G` correspond to inequalities formed from lower bounds, and the second column list which columns correspond to the matching upper bounds.

Imposing this complementarity constraint can substantially improve the numerical stability of the inner optimization.

**Example:** Suppose the original model has 5 moments:
* 1-2 are inequalities
* 3-5 are equalities, of which 3 and 4 involve a reduced-form parameter.

For projection inference,
* new moments 1 and 2 are old moments 1 and 2 (possibly relaxed using the upper limit of a one-sided CI if they involve a reduced-form parameter)
* new moments 3 and 4 are inequalities formed by inverting a CI for the reduced-form parameter in old moment 3
* new moments 5 and 6 are inequalities formed by inverting a CI for the reduced-form parameter in old moment 4
* old moment 5 becomes new moment 7.

In the `obj` representing the new model with 7 moments, set
* `d = 7`
* `inequality_index = [1, 2, 3, 4, 5, 6]`
* `complement_index = [3 4; 5 6]`

Refer to the replication files for the Chiappori et al. (2017) and Rust (1987) examples to see how to implement projection inference in practice.

## Local sensitivity

The local sensitivity measure at a particular parameter value `θ` is computed using the command
```
s = local_sensitivity(obj, θ)
```
Due to potential numerical instabilities involved when inverting large covariance matrices, accurate computation in models with many moments may require a large number of monte carlo draws.
